# -*- coding: utf-8 -*-
"""
Emails and Reports Helpers
~~~~~~~~~~~~~~~~~~~~~~~~~~

Functions to create emails and reports that are in common for both author_search and ref_search.
"""

import copy
import os
import re

import pandas

from . import ref_srch_emails_and_reports
from . import fileio


simple_publication_keywords_map = {"<abstract>":"abstract",
                                    "<conclusions>":"conclusions",
                                    "<copyrights>":"copyrights",
                                    "<DOI>":"doi",
                                    "<journal>":"journal",
                                    "<keywords>":"keywords",
                                    "<methods>":"methods",
                                    "<PMID>":"pubmed_id",
                                    "<results>":"results",
                                    "<title>":"title",
                                    "<PMCID>":"PMCID",}

pub_authors_keyword_map = {"<pub_author_first>":"firstname",
                           "<pub_author_last>":"lastname",
                           "<pub_author_initials>":"initials",
                           "<pub_author_collective>": "collectivename",
                           "<pub_author_affiliations>":"affiliation",
                           "<pub_author_ORCID>":"ORCID",
                           "<pub_author_id>":"author_id"}

references_keyword_map = {"<reference_citation>":"citation",
                          "<reference_title>":"title",
                          "<reference_PMID>":"pubmed_id",
                          "<reference_PMCID>":"PMCID",
                          "<reference_DOI>":"doi"}

publication_date_keywords_map = {"<publication_year>":"year",
                                 "<publication_month>":"month",
                                 "<publication_day>":"day"}

authors_keywords_map = {"<author_first>":"first_name",
                        "<author_last>":"last_name",
                        "<author_collective>": "collective_name",
                        "<author_name_search>":"pubmed_name_search",
                        "<author_email>":"email"}

pub_keywords = list(simple_publication_keywords_map.keys()) + \
               list(publication_date_keywords_map.keys()) + \
               ["<first_author>", "<last_author>", "<authors>", "<grants>", "<queried_sources>"]

## Ref search only.
tokenized_keywords_map = {"<tok_title>":"title", 
                          "<tok_DOI>":"DOI", 
                          "<tok_PMID>":"PMID"}



def _replace_keywords(template, publication_dict, config_dict, 
                     project_name="", author="", pub="", 
                     pub_author=None, reference=None, 
                     tokenized_citation=None, is_citation_in_prev_pubs=None):
    """Replace keywords in the values of the template dictionary.
    
    This is a convenience function to replace keywords as needed in strings. It is 
    setup so template is a dictionary for creating tabular reports from DataFrames. 
    So dict to DataFrame to csv or xlsx. To replace in a single string simply wrap 
    that single string in a dictionary with a single key and then pull it back out 
    of that dictionary from that key.
    
    This was merged to handle both reference search and author search. Previously there 
    were 2 replace_keywords functions, one for each search type.
    
    Args:
        template (dict): keys are column names and values are what the elements of the column should be.
        publication_dict (dict): keys and values match the publications JSON file.
        config_dict (dict): keys and values match the project tracking configuration JSON file.
        project_name (str): the name of the project to replace.
        author (str): the key to the author in config_dict["Authors"].
        pub (str): the key to the pub in publication_dict.
        pub_author (dict|None): an author in pub. None means no author.
        reference (dict|None): a reference for pub. None means no reference.
        tokenized_citation (dict): a tokenized citation from the reference for the publication.
        is_citation_in_prev_pubs (bool or None): whether this publication is in the previous publications or not. If None then it isn't applicable.
        
    Returns:
        template_copy (dict): template with the keywords replaced in its values.
    """
    
    template_copy = copy.deepcopy(template)
    
    for key in template_copy:
        
        ## Project keywords
        if project_name:
            template_copy[key] = template_copy[key].replace("<project_name>", project_name)
        
        ## Authors keywords
        if author:
            for keyword, auth_key in authors_keywords_map.items():
                template_copy[key] = template_copy[key].replace(keyword, str(config_dict["Authors"][author][auth_key]) if auth_key in config_dict["Authors"][author] else "None")
        
        ## Publication keywords
        if pub:
            for keyword, pub_key in simple_publication_keywords_map.items():
                template_copy[key] = template_copy[key].replace(keyword, str(publication_dict[pub][pub_key]))
                
            ## build first and last author
            if publication_dict[pub]["authors"]:
                first_author_attributes = publication_dict[pub]["authors"][0]
                if "collectivename" in first_author_attributes:
                    first_author = first_author_attributes["collectivename"]
                else:
                    first_author = str(first_author_attributes["lastname"]) + ", " + str(first_author_attributes["firstname"])
                
                last_author_attributes = publication_dict[pub]["authors"][-1]
                if "collectivename" in last_author_attributes:
                    last_author = last_author_attributes["collectivename"]
                else:
                    last_author = str(last_author_attributes["lastname"]) + ", " + str(last_author_attributes["firstname"])
                
                # first_author = str(publication_dict[pub]["authors"][0]["lastname"]) + ", " + str(publication_dict[pub]["authors"][0]["firstname"])
                template_copy[key] = template_copy[key].replace("<first_author>", first_author)
                
                # last_author = str(publication_dict[pub]["authors"][-1]["lastname"]) + ", " + str(publication_dict[pub]["authors"][-1]["firstname"])
                template_copy[key] = template_copy[key].replace("<last_author>", last_author)
                
                authors = []
                for author_attributes in publication_dict[pub]["authors"]:
                    if "collectivename" in author_attributes:
                        authors.append(author_attributes["collectivename"])
                    else:
                        authors.append(str(author_attributes["firstname"]) + " " + str(author_attributes["lastname"]))
                authors = ", ".join(authors)
            
                # authors = ", ".join([str(author["firstname"]) + " " + str(author["lastname"]) for author in publication_dict[pub]["authors"]])
                template_copy[key] = template_copy[key].replace("<authors>", authors)
            else:
                first_author = "No Authors"
                last_author = "No Authors"
                authors = "No Authors"
            
            grants = ", ".join(publication_dict[pub]["grants"]) if publication_dict[pub]["grants"] else "None Found"
            template_copy[key] = template_copy[key].replace("<grants>", grants)
            
            queried_sources = ", ".join(publication_dict[pub]["queried_sources"])
            template_copy[key] = template_copy[key].replace("<queried_sources>", queried_sources)
            
            for keyword, date_key in publication_date_keywords_map.items():
                template_copy[key] = template_copy[key].replace(keyword, str(publication_dict[pub]["publication_date"][date_key]))
        
        ## Pub authors keywords
        if pub_author:
            for keyword, pub_author_key in pub_authors_keyword_map.items():
                template_copy[key] = template_copy[key].replace(keyword, str(pub_author[pub_author_key]) if pub_author_key in pub_author else "None")
        
        ## references keywords
        if reference:
            for keyword, reference_key in references_keyword_map.items():
                template_copy[key] = template_copy[key].replace(keyword, str(reference[reference_key]))
        
        ## tokenized keywords
        if tokenized_citation:
            for keyword, tok_key in tokenized_keywords_map.items():
                replacement = str(tokenized_citation[tok_key])
                if not replacement:
                    replacement = "None"
                template_copy[key] = template_copy[key].replace(keyword, replacement)
                
            tok_authors = ref_srch_emails_and_reports.convert_tokenized_authors_to_str(tokenized_citation["authors"])
            template_copy[key] = template_copy[key].replace("<tok_authors>", tok_authors)
            
            if tokenized_citation["reference_line"]:
                pretty_print = tokenized_citation["reference_line"].split("\n")
                pretty_print = " ".join([line.strip() for line in pretty_print])
                template_copy[key] = template_copy[key].replace("<ref_line>", pretty_print)
            else:
                template_copy[key] = template_copy[key].replace("<ref_line>", "N/A")
            
            if type(is_citation_in_prev_pubs) == bool:
                template_copy[key] = template_copy[key].replace("<is_in_comparison_file>", str(is_citation_in_prev_pubs))
            else:
                template_copy[key] = template_copy[key].replace("<is_in_comparison_file>", "N/A")
            
    return template_copy



def _replace_pub_author_and_reference_loops(publication_dict, pub, string_to_modify, pub_author_template, reference_template):
    """Replace the pub_author and reference loops in string_to_modify with appropriate pub data.
    
    Args:
        publication_dict (dict): keys and values match the publications JSON file.
        pub (str): the key to the pub in publication_dict.
        string_to_modify (str): the string to place the pub_author and reference information into.
        pub_author_template (str): string that serves as a template for how to build the information about each pub_author.
        reference_template (str): string that serves as a template for how to build the information about each reference.
    
    Returns:
        string_to_modify (str): the input string_to_modify with pub_author and reference loops replaced.
    """
    
    pub_authors = ""
    for pub_author in publication_dict[pub]["authors"]:
        pub_author_template_copy = _replace_keywords({"1":pub_author_template}, publication_dict, {}, pub_author=pub_author)["1"]
        pub_authors += pub_author_template_copy
        
    string_to_modify = re.sub(r"(?s)<pub_author_loop>.*</pub_author_loop>", pub_authors, string_to_modify)
    
    references = ""
    for reference in publication_dict[pub]["references"]:
        reference_template_copy = _replace_keywords({"1":reference_template}, publication_dict, {}, reference=reference)["1"]
        references += reference_template_copy
    
    if not references:
        references = "None"
    string_to_modify = re.sub(r"(?s)<reference_loop>.*</reference_loop>", references, string_to_modify)
    
    return string_to_modify



def _save_rows_to_file(rows, filename, sort, column_order, file_format, separator, save_dir_name):
    """Turn a list of dicts into a DataFrame and save it to file.
    
    Args:
        rows (list[dict]): list of dictionaries to save.
        filename (str): filename to save the DataFrame as, '.csv' or '.xlsx' will be added as needed.
        sort (list[str]): list of column names to sort the DataFrame by, passes into DataFrame.sort_values().
        column_order (list[str]): list of column names to reorder the DataFrame columns.
        file_format (str): either 'csv' or 'xlsx' to save in CSV or Excel format, respectively.
        separator (str): the separator to use if file_format is 'csv'.
        save_dir_name (str): the directory name to save the file in.
    
    Returns:
        report (str): if file_format is 'csv', then the DataFrame in CSV form, else the path to the Excel file.
        filename (str): the filename the DataFrame was ultimately saved under.
    """
    
    report = ""            
    if rows:
        df = pandas.DataFrame(rows)
        if sort:
            df = df.sort_values(by=sort)
        df = df.drop_duplicates()
        df = df[column_order]
        
        if file_format == "csv":
            report = df.to_csv(index=False, sep=separator, lineterminator="\n")
            fileio.save_string_to_file(save_dir_name, filename, report)
        else:
            ## If the file extension isn't .xlsx then there will be an error, so force it.
            extension = os.path.splitext(filename)[1][1:].lower()
            if not extension == "xlsx":
                filename += ".xlsx"
            
            report = os.path.join(save_dir_name, filename)
            df.to_excel(report, index=False)
    
    return report, filename



def _build_pub_author_and_reference_rows(publication_dict, config_dict, 
                                         has_pub_author_keywords, has_reference_keywords,
                                         row_template, 
                                         project_name="", author="", pub="", 
                                         tokenized_citation=None, is_citation_in_prev_pubs=None):
    """Build rows for each pub_author and reference.
    
    Args:
        publication_dict (dict): keys and values match the publications JSON file.
        config_dict (dict): keys and values match the project tracking configuration JSON file.
        has_pub_author_keywords (bool): if True, then row_template has keywords to replace that are attributes to publication authors.
        has_reference_keywords (bool): if True, then row_template has keywords to replace that are attributes to publication references.
        row_template (dict): keys are column names and values are what the elements of the column should be.
        project_name (str): the name of the project to replace.
        author (str): the key to the author in config_dict["Authors"].
        pub (str): the key to the pub in publication_dict.
        tokenized_citation (dict|None): a tokenized citation from the reference for the publication.
        is_citation_in_prev_pubs (bool|None): whether this publication is in the previous publications or not. If None then it isn't applicable.
    
    Returns:
        rows (list): list of dictionaries meant to eventually be turned into a pandas DataFrame.
    """
    
    rows = []
    ## If references or authors is an empty list then you get unexpected behavior where pubs just won't show up, 
    ## so look for it before hand and just give it a null dictionary if it is empty.
    if not (references := publication_dict[pub]["references"]):
        references = [{
                        "PMCID": None,
                        "citation": None,
                        "doi": None,
                        "pubmed_id": None,
                        "title": None
                      }]
    ## There should always be at least 1 author since this is specific to author search, but the code is here for completeness.
    if not (pub_authors := publication_dict[pub]["authors"]):
        pub_authors = [{
                        "ORCID": None,
                        "affiliation": None,
                        "author_id": None,
                        "firstname": None,
                        "initials": None,
                        "lastname": None
                      }]
    
    if has_pub_author_keywords and has_reference_keywords:
        for pub_author in pub_authors:
            for reference in references:
                rows.append(_replace_keywords(row_template, 
                                              publication_dict, 
                                              config_dict, 
                                              project_name, 
                                              author,
                                              pub, 
                                              pub_author, 
                                              reference,
                                              tokenized_citation,
                                              is_citation_in_prev_pubs))
    
    elif has_pub_author_keywords:
        for pub_author in pub_authors:
            rows.append(_replace_keywords(row_template, 
                                          publication_dict, 
                                          config_dict, 
                                          project_name, 
                                          author, 
                                          pub, 
                                          pub_author,
                                          None,
                                          tokenized_citation,
                                          is_citation_in_prev_pubs))
    
    elif has_reference_keywords:
        for reference in references:
            rows.append(_replace_keywords(row_template, 
                                          publication_dict, 
                                          config_dict, 
                                          project_name, 
                                          author, 
                                          pub, 
                                          None, 
                                          reference,
                                          tokenized_citation,
                                          is_citation_in_prev_pubs))
            
    else:
        rows.append(_replace_keywords(row_template, 
                                      publication_dict, 
                                      config_dict, 
                                      project_name, 
                                      author, 
                                      pub,
                                      None,
                                      None,
                                      tokenized_citation,
                                      is_citation_in_prev_pubs))
    
    return rows



# -*- coding: utf-8 -*-


import os
import requests
import re
import datetime
import xml.etree.ElementTree as ET
import json
import copy

import pytest
import shutil
import pymed

from academic_tracker.athr_srch_modularized import input_reading_and_checking, generate_internal_data_and_check_authors
from academic_tracker.athr_srch_modularized import build_publication_dict, save_and_send_reports_and_emails
from academic_tracker.fileio import load_json, read_text_from_txt
from academic_tracker.athr_srch_webio import search_PubMed_for_pubs, search_ORCID_for_pubs, search_Google_Scholar_for_pubs, search_Crossref_for_pubs


@pytest.fixture(autouse=True)
def disable_network_calls(monkeypatch):
    def stunted_get():
        raise RuntimeError("Network access not allowed during testing!")
    monkeypatch.setattr(requests, "get", lambda *args, **kwargs: stunted_get())
    


@pytest.fixture
def config_dict():
    return load_json(os.path.join("tests", "testing_files", "config_truncated.json"))

@pytest.fixture
def config_dict_Hunter_only():
    return load_json(os.path.join("tests", "testing_files", "config_Hunter_only.json"))

@pytest.fixture
def original_queries():
    query_json = load_json(os.path.join("tests", "testing_files", "all_queries.json"))
    ## Convert PubMed dictionaries back to articles class.
    for author, pub_list in query_json["PubMed"].items():
        new_list = []
        for pub in pub_list:
            new_list.append(pymed.article.PubMedArticle(ET.fromstring(pub["xml"])))
        query_json["PubMed"][author] = new_list
    return query_json





def test_input_reading_and_checking(config_dict):
    config_json_filepath = os.path.join("tests", "testing_files", "config_truncated.json")
    
    expected_config_dict = config_dict
    
    actual_config_dict = input_reading_and_checking(config_json_filepath, False, False, False, False)
    
    assert expected_config_dict == actual_config_dict
    

def test_input_reading_and_checking_no_ORCID():
    config_json_filepath = os.path.join("tests", "testing_files", "config_truncated_noORCID.json")
    
    expected_config_dict = load_json(os.path.join("tests", "testing_files", "config_truncated_noORCID.json"))
    
    actual_config_dict = input_reading_and_checking(config_json_filepath, False, False, False, False)
    
    assert expected_config_dict == actual_config_dict
    

def test_input_reading_and_checking_no_PubMed():
    config_json_filepath = os.path.join("tests", "testing_files", "config_truncated_noPubMed.json")
    
    expected_config_dict = load_json(os.path.join("tests", "testing_files", "config_truncated_noPubMed.json"))
    
    actual_config_dict = input_reading_and_checking(config_json_filepath, False, False, False, False)
    
    assert expected_config_dict == actual_config_dict
    
    
def test_input_reading_and_checking_no_Crossref():
    config_json_filepath = os.path.join("tests", "testing_files", "config_truncated_noCrossref.json")
        
    with pytest.raises(BaseException):
        actual_config_dict = input_reading_and_checking(config_json_filepath, False, False, False, False)
        

def test_input_reading_and_checking_no_Crossref_noGS():
    config_json_filepath = os.path.join("tests", "testing_files", "config_truncated_noCrossref.json")
        
    expected_config_dict = load_json(os.path.join("tests", "testing_files", "config_truncated_noCrossref.json"))
    
    actual_config_dict = input_reading_and_checking(config_json_filepath, False, True, False, False)
    
    assert expected_config_dict == actual_config_dict
    



def test_generate_internal_data_and_check_authors(config_dict, capsys):
    expected_authors_by_project_dict = load_json(os.path.join("tests", "testing_files", "authors_by_project_dict_truncated.json"))
    expected_config_dict = load_json(os.path.join("tests", "testing_files", "config_truncated_authors_adjusted.json"))
    
    actual_authors_by_project_dict, actual_config_dict = generate_internal_data_and_check_authors(config_dict)
    # with open(os.path.join("tests", "testing_files", "config_truncated_authors_adjusted_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(actual_config_dict, indent=2, sort_keys=True))
    
    captured = capsys.readouterr()
    
    assert len(captured.out) > 0
    assert expected_authors_by_project_dict == actual_authors_by_project_dict
    assert expected_config_dict == actual_config_dict



def test_build_publication_dict_all_sources(mocker, config_dict_Hunter_only, original_queries):
    ## Code to run and save intermediate running_pubs. This is slow, so they are saved once and read in later.
    # running_pubs = {}
    # running_pubs1, _ = search_PubMed_for_pubs(running_pubs, config_dict_Hunter_only["Authors"], "asdf", original_queries["PubMed"])
    # running_pubs2, _ = search_ORCID_for_pubs(copy.deepcopy(running_pubs1), "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
    # running_pubs3, _ = search_Google_Scholar_for_pubs(copy.deepcopy(running_pubs2), config_dict_Hunter_only["Authors"], "asdf", original_queries["Google Scholar"])
    # running_pubs4, _ = search_Crossref_for_pubs(copy.deepcopy(running_pubs3), config_dict_Hunter_only["Authors"], "asdf", original_queries["Crossref"])
    
    # running_pubs5, _ = search_PubMed_for_pubs(copy.deepcopy(running_pubs4), config_dict_Hunter_only["Authors"], "asdf", original_queries["PubMed"])
    # running_pubs6, _ = search_ORCID_for_pubs(copy.deepcopy(running_pubs5), "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
    # running_pubs7, _ = search_Google_Scholar_for_pubs(copy.deepcopy(running_pubs6), config_dict_Hunter_only["Authors"], "asdf", original_queries["Google Scholar"])
    # running_pubs8, _ = search_Crossref_for_pubs(copy.deepcopy(running_pubs7), config_dict_Hunter_only["Authors"], "asdf", original_queries["Crossref"])
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs1.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs1, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs2.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs2, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs3.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs3, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs4.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs4, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs5.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs5, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs6.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs6, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs7.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs7, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs8.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs8, indent=2, sort_keys=True))
    
    
    running_pubs1 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs1.json"))
    running_pubs2 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs2.json"))
    running_pubs3 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs3.json"))
    running_pubs4 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs4.json"))
    running_pubs5 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs5.json"))
    running_pubs6 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs6.json"))
    running_pubs7 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs7.json"))
    running_pubs8 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs8.json"))
    
            
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_PubMed_for_pubs", 
                  side_effect=[(running_pubs1, original_queries["PubMed"]), (running_pubs5, original_queries["PubMed"])])
    
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_ORCID_for_pubs", 
                  side_effect=[(running_pubs2, original_queries["ORCID"]), (running_pubs6, original_queries["ORCID"])])
    
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_Google_Scholar_for_pubs", 
                  side_effect=[(running_pubs3, original_queries["Google Scholar"]), (running_pubs7, original_queries["Google Scholar"])])
    
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_Crossref_for_pubs", 
                  side_effect=[(running_pubs4, original_queries["Crossref"]), (running_pubs8, original_queries["Crossref"])])
    
    
    actual_publication_dict, _ = build_publication_dict(config_dict_Hunter_only, {}, False, False, False, False)
    
    ## Have to sort to match how it is saved as JSON.
    actual_publication_dict = {key:{key2:value2 for key2, value2 in sorted(value.items())} for key, value in sorted(actual_publication_dict.items())}
    
    expected_pub_dict = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "publication_dict.json"))
        
    assert expected_pub_dict == actual_publication_dict



def test_build_publication_dict_no_ORCID(mocker, config_dict_Hunter_only, original_queries):
    ## Code to run and save intermediate running_pubs. This is slow, so they are saved once and read in later.
    # running_pubs = {}
    # running_pubs1, _ = search_PubMed_for_pubs(running_pubs, config_dict_Hunter_only["Authors"], "asdf", original_queries["PubMed"])
    # running_pubs2, _ = search_Google_Scholar_for_pubs(copy.deepcopy(running_pubs1), config_dict_Hunter_only["Authors"], "asdf", original_queries["Google Scholar"])
    # running_pubs3, _ = search_Crossref_for_pubs(copy.deepcopy(running_pubs2), config_dict_Hunter_only["Authors"], "asdf", original_queries["Crossref"])
    
    # running_pubs4, _ = search_PubMed_for_pubs(copy.deepcopy(running_pubs3), config_dict_Hunter_only["Authors"], "asdf", original_queries["PubMed"])
    # running_pubs5, _ = search_Google_Scholar_for_pubs(copy.deepcopy(running_pubs4), config_dict_Hunter_only["Authors"], "asdf", original_queries["Google Scholar"])
    # running_pubs6, _ = search_Crossref_for_pubs(copy.deepcopy(running_pubs5), config_dict_Hunter_only["Authors"], "asdf", original_queries["Crossref"])
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_ORCID", "running_pubs1.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs1, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_ORCID", "running_pubs2.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs2, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_ORCID", "running_pubs3.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs3, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_ORCID", "running_pubs4.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs4, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_ORCID", "running_pubs5.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs5, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_ORCID", "running_pubs6.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs6, indent=2, sort_keys=True))
    
    
    running_pubs1 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_ORCID", "running_pubs1.json"))
    running_pubs2 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_ORCID", "running_pubs2.json"))
    running_pubs3 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_ORCID", "running_pubs3.json"))
    running_pubs4 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_ORCID", "running_pubs4.json"))
    running_pubs5 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_ORCID", "running_pubs5.json"))
    running_pubs6 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_ORCID", "running_pubs6.json"))
    
            
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_PubMed_for_pubs", 
                  side_effect=[(running_pubs1, original_queries["PubMed"]), (running_pubs4, original_queries["PubMed"])])
        
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_Google_Scholar_for_pubs", 
                  side_effect=[(running_pubs2, original_queries["Google Scholar"]), (running_pubs5, original_queries["Google Scholar"])])
    
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_Crossref_for_pubs", 
                  side_effect=[(running_pubs3, original_queries["Crossref"]), (running_pubs6, original_queries["Crossref"])])
    
    
    actual_publication_dict, _ = build_publication_dict(config_dict_Hunter_only, {}, True, False, False, False)
    
    ## Have to sort to match how it is saved as JSON.
    actual_publication_dict = {key:{key2:value2 for key2, value2 in sorted(value.items())} for key, value in sorted(actual_publication_dict.items())}
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_ORCID", "publication_dict.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(actual_publication_dict, indent=2, sort_keys=True))
    
    expected_pub_dict = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_ORCID", "publication_dict.json"))
    
    assert expected_pub_dict == actual_publication_dict
    


def test_build_publication_dict_no_Crossref(mocker, config_dict_Hunter_only, original_queries):
    ## Code to run and save intermediate running_pubs. This is slow, so they are saved once and read in later.
    # running_pubs = {}
    # running_pubs1, _ = search_PubMed_for_pubs(running_pubs, config_dict_Hunter_only["Authors"], "asdf", original_queries["PubMed"])
    # running_pubs2, _ = search_ORCID_for_pubs(copy.deepcopy(running_pubs1), "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
    # running_pubs3, _ = search_Google_Scholar_for_pubs(copy.deepcopy(running_pubs2), config_dict_Hunter_only["Authors"], "asdf", original_queries["Google Scholar"])
    
    # running_pubs4, _ = search_PubMed_for_pubs(copy.deepcopy(running_pubs3), config_dict_Hunter_only["Authors"], "asdf", original_queries["PubMed"])
    # running_pubs5, _ = search_ORCID_for_pubs(copy.deepcopy(running_pubs4), "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
    # running_pubs6, _ = search_Google_Scholar_for_pubs(copy.deepcopy(running_pubs5), config_dict_Hunter_only["Authors"], "asdf", original_queries["Google Scholar"])
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Crossref", "running_pubs1.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs1, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Crossref", "running_pubs2.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs2, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Crossref", "running_pubs3.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs3, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Crossref", "running_pubs4.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs4, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Crossref", "running_pubs5.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs5, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Crossref", "running_pubs6.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs6, indent=2, sort_keys=True))
    
    
    running_pubs1 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Crossref", "running_pubs1.json"))
    running_pubs2 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Crossref", "running_pubs2.json"))
    running_pubs3 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Crossref", "running_pubs3.json"))
    running_pubs4 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Crossref", "running_pubs4.json"))
    running_pubs5 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Crossref", "running_pubs5.json"))
    running_pubs6 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Crossref", "running_pubs6.json"))
    
            
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_PubMed_for_pubs", 
                  side_effect=[(running_pubs1, original_queries["PubMed"]), (running_pubs4, original_queries["PubMed"])])
        
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_ORCID_for_pubs", 
                  side_effect=[(running_pubs2, original_queries["ORCID"]), (running_pubs5, original_queries["ORCID"])])
    
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_Google_Scholar_for_pubs", 
                  side_effect=[(running_pubs3, original_queries["Google Scholar"]), (running_pubs6, original_queries["Google Scholar"])])
    
    
    actual_publication_dict, _ = build_publication_dict(config_dict_Hunter_only, {}, False, False, True, False)
    
    ## Have to sort to match how it is saved as JSON.
    actual_publication_dict = {key:{key2:value2 for key2, value2 in sorted(value.items())} for key, value in sorted(actual_publication_dict.items())}
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Crossref", "publication_dict.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(actual_publication_dict, indent=2, sort_keys=True))
    
    expected_pub_dict = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Crossref", "publication_dict.json"))
    
    assert expected_pub_dict == actual_publication_dict
    


def test_build_publication_dict_no_Google_Scholar(mocker, config_dict_Hunter_only, original_queries):
    ## Code to run and save intermediate running_pubs. This is slow, so they are saved once and read in later.
    # running_pubs = {}
    # running_pubs1, _ = search_PubMed_for_pubs(running_pubs, config_dict_Hunter_only["Authors"], "asdf", original_queries["PubMed"])
    # running_pubs2, _ = search_ORCID_for_pubs(copy.deepcopy(running_pubs1), "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
    # running_pubs3, _ = search_Crossref_for_pubs(copy.deepcopy(running_pubs2), config_dict_Hunter_only["Authors"], "asdf", original_queries["Crossref"])
    
    # running_pubs4, _ = search_PubMed_for_pubs(copy.deepcopy(running_pubs3), config_dict_Hunter_only["Authors"], "asdf", original_queries["PubMed"])
    # running_pubs5, _ = search_ORCID_for_pubs(copy.deepcopy(running_pubs4), "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
    # running_pubs6, _ = search_Crossref_for_pubs(copy.deepcopy(running_pubs5), config_dict_Hunter_only["Authors"], "asdf", original_queries["Crossref"])
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Google_Scholar", "running_pubs1.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs1, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Google_Scholar", "running_pubs2.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs2, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Google_Scholar", "running_pubs3.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs3, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Google_Scholar", "running_pubs4.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs4, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Google_Scholar", "running_pubs5.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs5, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Google_Scholar", "running_pubs6.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs6, indent=2, sort_keys=True))
    
    
    running_pubs1 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Google_Scholar", "running_pubs1.json"))
    running_pubs2 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Google_Scholar", "running_pubs2.json"))
    running_pubs3 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Google_Scholar", "running_pubs3.json"))
    running_pubs4 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Google_Scholar", "running_pubs4.json"))
    running_pubs5 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Google_Scholar", "running_pubs5.json"))
    running_pubs6 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Google_Scholar", "running_pubs6.json"))
    
            
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_PubMed_for_pubs", 
                  side_effect=[(running_pubs1, original_queries["PubMed"]), (running_pubs4, original_queries["PubMed"])])
        
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_ORCID_for_pubs", 
                  side_effect=[(running_pubs2, original_queries["ORCID"]), (running_pubs5, original_queries["ORCID"])])
    
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_Crossref_for_pubs", 
                  side_effect=[(running_pubs3, original_queries["Crossref"]), (running_pubs6, original_queries["Crossref"])])
    
    
    actual_publication_dict, _ = build_publication_dict(config_dict_Hunter_only, {}, False, True, False, False)
    
    ## Have to sort to match how it is saved as JSON.
    actual_publication_dict = {key:{key2:value2 for key2, value2 in sorted(value.items())} for key, value in sorted(actual_publication_dict.items())}
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Google_Scholar", "publication_dict.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(actual_publication_dict, indent=2, sort_keys=True))
    
    expected_pub_dict = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_Google_Scholar", "publication_dict.json"))
    
    assert expected_pub_dict == actual_publication_dict
    
 
    
def test_build_publication_dict_no_PubMed(mocker, config_dict_Hunter_only, original_queries):
    ## Code to run and save intermediate running_pubs. This is slow, so they are saved once and read in later.
    # running_pubs = {}
    # running_pubs1, _ = search_ORCID_for_pubs(running_pubs, "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
    # running_pubs2, _ = search_Google_Scholar_for_pubs(copy.deepcopy(running_pubs1), config_dict_Hunter_only["Authors"], "asdf", original_queries["Google Scholar"])
    # running_pubs3, _ = search_Crossref_for_pubs(copy.deepcopy(running_pubs2), config_dict_Hunter_only["Authors"], "asdf", original_queries["Crossref"])
    
    # running_pubs4, _ = search_ORCID_for_pubs(copy.deepcopy(running_pubs3), "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
    # running_pubs5, _ = search_Google_Scholar_for_pubs(copy.deepcopy(running_pubs4), config_dict_Hunter_only["Authors"], "asdf", original_queries["Google Scholar"])
    # running_pubs6, _ = search_Crossref_for_pubs(copy.deepcopy(running_pubs5), config_dict_Hunter_only["Authors"], "asdf", original_queries["Crossref"])
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_PubMed", "running_pubs1.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs1, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_PubMed", "running_pubs2.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs2, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_PubMed", "running_pubs3.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs3, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_PubMed", "running_pubs4.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs4, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_PubMed", "running_pubs5.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs5, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_PubMed", "running_pubs6.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs6, indent=2, sort_keys=True))
    
    
    running_pubs1 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_PubMed", "running_pubs1.json"))
    running_pubs2 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_PubMed", "running_pubs2.json"))
    running_pubs3 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_PubMed", "running_pubs3.json"))
    running_pubs4 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_PubMed", "running_pubs4.json"))
    running_pubs5 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_PubMed", "running_pubs5.json"))
    running_pubs6 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_PubMed", "running_pubs6.json"))
    
            
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_ORCID_for_pubs", 
                  side_effect=[(running_pubs1, original_queries["ORCID"]), (running_pubs4, original_queries["ORCID"])])
    
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_Google_Scholar_for_pubs", 
                  side_effect=[(running_pubs2, original_queries["Google Scholar"]), (running_pubs5, original_queries["Google Scholar"])])
    
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_Crossref_for_pubs", 
                  side_effect=[(running_pubs3, original_queries["Crossref"]), (running_pubs6, original_queries["Crossref"])])
    
    
    actual_publication_dict, _ = build_publication_dict(config_dict_Hunter_only, {}, False, False, False, True)
    
    ## Have to sort to match how it is saved as JSON.
    actual_publication_dict = {key:{key2:value2 for key2, value2 in sorted(value.items())} for key, value in sorted(actual_publication_dict.items())}
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_PubMed", "publication_dict.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(actual_publication_dict, indent=2, sort_keys=True))
    
    expected_pub_dict = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_PubMed", "publication_dict.json"))
    
    assert expected_pub_dict == actual_publication_dict
    

    
def test_build_publication_dict_duplicates(mocker, config_dict_Hunter_only, original_queries):
    expected_pub_dict = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "publication_dict.json"))
    pub_to_duplicate = ""
    for pub_id in expected_pub_dict:
        pub_to_duplicate = pub_id
        break
    prev_pubs = {pub_to_duplicate:expected_pub_dict[pub_to_duplicate]}
    
    
    running_pubs1 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs1.json"))
    running_pubs2 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs2.json"))
    running_pubs3 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs3.json"))
    running_pubs4 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs4.json"))
    running_pubs5 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs5.json"))
    running_pubs6 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs6.json"))
    running_pubs7 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs7.json"))
    running_pubs8 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs8.json"))
    
            
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_PubMed_for_pubs", 
                  side_effect=[(running_pubs1, original_queries["PubMed"]), (running_pubs5, original_queries["PubMed"])])
    
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_ORCID_for_pubs", 
                  side_effect=[(running_pubs2, original_queries["ORCID"]), (running_pubs6, original_queries["ORCID"])])
    
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_Google_Scholar_for_pubs", 
                  side_effect=[(running_pubs3, original_queries["Google Scholar"]), (running_pubs7, original_queries["Google Scholar"])])
    
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_Crossref_for_pubs", 
                  side_effect=[(running_pubs4, original_queries["Crossref"]), (running_pubs8, original_queries["Crossref"])])
    
    actual_publication_dict, _ = build_publication_dict(config_dict_Hunter_only, prev_pubs, False, False, False, False)
    
    ## Have to sort to match how it is saved as JSON.
    actual_publication_dict = {key:{key2:value2 for key2, value2 in sorted(value.items())} for key, value in sorted(actual_publication_dict.items())}
    
    del expected_pub_dict[pub_to_duplicate]
    
    assert expected_pub_dict == actual_publication_dict



def test_build_publication_dict_no_pubs_found(mocker, config_dict_Hunter_only, capsys):
    # config_dict = load_json(os.path.join("tests", "testing_files", "config_truncated.json"))
           
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_PubMed_for_pubs", 
                  side_effect=[({}, {}), ({}, {})])
    
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_ORCID_for_pubs", 
                  side_effect=[({}, {}), ({}, {})])
    
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_Google_Scholar_for_pubs", 
                  side_effect=[({}, {}), ({}, {})])
    
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_webio.search_Crossref_for_pubs", 
                  side_effect=[({}, {}), ({}, {})])
    
    with pytest.raises(SystemExit):
        actual_publication_dict, _ = build_publication_dict(config_dict_Hunter_only, {}, False, False, False, False)
    
    captured = capsys.readouterr()
    expected_message = "No new publications found." + "\n"
    assert expected_message in captured.out
    



def test_save_and_send_reports_and_emails_no_email(config_dict, mocker):
    
    def emails(*args, **kwargs):
        return {"creation_date":"asdf", "emails":[]}
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_emails_and_reports.create_project_reports_and_emails", emails)
    
    pub_dict = load_json(os.path.join("tests", "testing_files", "publication_dict_truncated.json"))
    authors_by_project_dict = load_json(os.path.join("tests", "testing_files", "authors_by_project_dict_truncated.json"))
    
    config_dict["summary_report"] = {}
    config_dict["summary_report"]["template"] = read_text_from_txt(os.path.join("tests", "testing_files", "athr_srch_build_loop_template_string.txt"))
    
    save_dir = save_and_send_reports_and_emails(authors_by_project_dict, pub_dict, config_dict, True)
    # with open(os.path.join("tests", "testing_files", "athr_srch_summary_report_custom_template_new.txt"), 'wb') as outFile:
    #     outFile.write(read_text_from_txt(os.path.join(save_dir, "summary_report.txt")).encode("utf-8"))
    
    
    assert os.path.exists(os.path.join(save_dir, "summary_report.txt"))
    assert read_text_from_txt(os.path.join("tests", "testing_files", "athr_srch_summary_report_custom_template.txt")) == read_text_from_txt(os.path.join(save_dir, "summary_report.txt"))
    assert not os.path.exists(os.path.join(save_dir, "emails.json"))



def test_save_and_send_reports_and_emails_with_email(config_dict, mocker):
    
    def emails(*args, **kwargs):
        return {"creation_date":"asdf", "emails":[]}
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_emails_and_reports.create_project_reports_and_emails", emails)
    
    pub_dict = load_json(os.path.join("tests", "testing_files", "publication_dict_truncated.json"))
    authors_by_project_dict = load_json(os.path.join("tests", "testing_files", "authors_by_project_dict_truncated.json"))
    
    config_dict["summary_report"] = {}
    config_dict["summary_report"]["template"] = read_text_from_txt(os.path.join("tests", "testing_files", "athr_srch_build_loop_template_string.txt"))
    config_dict["summary_report"]["from_email"] = "ptth222@uky.edu"
    config_dict["summary_report"]["to_email"] = ["ptth222@uky.edu"]
    config_dict["summary_report"]["cc_email"] = ["ptth222@uky.edu"]
    config_dict["summary_report"]["email_body"] = "Body"
    config_dict["summary_report"]["email_subject"] = "Subject"
    
    save_dir = save_and_send_reports_and_emails(authors_by_project_dict, pub_dict, config_dict, True)
    
    assert os.path.exists(os.path.join(save_dir, "summary_report.txt"))
    assert read_text_from_txt(os.path.join("tests", "testing_files", "athr_srch_summary_report_custom_template.txt")) == read_text_from_txt(os.path.join(save_dir, "summary_report.txt"))
    assert os.path.exists(os.path.join(save_dir, "emails.json"))



def test_save_and_send_reports_and_emails_default_template(config_dict, mocker):
    
    def emails(*args, **kwargs):
        return {"creation_date":"asdf", "emails":[]}
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_emails_and_reports.create_project_reports_and_emails", emails)
    
    pub_dict = load_json(os.path.join("tests", "testing_files", "publication_dict_truncated.json"))
    ## Add Travis to a publication so we can test that multiple authors get reported.
    pub_dict["https://doi.org/10.1038/s41597-023-02277-x"]["authors"][1]["author_id"] = "Travis Thompson"
    authors_by_project_dict = load_json(os.path.join("tests", "testing_files", "authors_by_project_dict_truncated.json"))
    
    config_dict["summary_report"] = {}
    
    save_dir = save_and_send_reports_and_emails(authors_by_project_dict, pub_dict, config_dict, True)
    # with open(os.path.join("tests", "testing_files", "athr_srch_summary_report_new.txt"), 'wb') as outFile:
    #     outFile.write(read_text_from_txt(os.path.join(save_dir, "summary_report.txt")).encode("utf-8"))
    
    assert os.path.exists(os.path.join(save_dir, "summary_report.txt"))
    assert read_text_from_txt(os.path.join("tests", "testing_files", "athr_srch_summary_report.txt")) == read_text_from_txt(os.path.join(save_dir, "summary_report.txt"))
    assert not os.path.exists(os.path.join(save_dir, "emails.json"))


def test_save_and_send_reports_and_emails_collab_reports(config_dict, mocker):
    
    def emails(*args, **kwargs):
        return {"creation_date":"asdf", "emails":[]}
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_emails_and_reports.create_project_reports_and_emails", emails)
    
    pub_dict = load_json(os.path.join("tests", "testing_files", "publication_dict_truncated.json"))
    authors_by_project_dict = load_json(os.path.join("tests", "testing_files", "authors_by_project_dict_truncated.json"))
    
    config_dict["Authors"]["Hunter Moseley"]["collaborator_report"] = {"from_email":"ptth222@uky.edu", 
                                                                    "email_body":"asdf",
                                                                    "email_subject":"asdf",
                                                                    }
    
    save_dir = save_and_send_reports_and_emails(authors_by_project_dict, pub_dict, config_dict, True)
    
    assert os.path.exists(os.path.join(save_dir, "Hunter Moseley_collaborators.csv"))
    assert os.path.exists(os.path.join(save_dir, "emails.json"))
    
    
def test_save_and_send_reports_and_emails_tabular_summary(config_dict, mocker):
    
    def emails(*args, **kwargs):
        return {"creation_date":"asdf", "emails":[]}
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_emails_and_reports.create_project_reports_and_emails", emails)
    
    pub_dict = load_json(os.path.join("tests", "testing_files", "publication_dict_truncated.json"))
    authors_by_project_dict = load_json(os.path.join("tests", "testing_files", "authors_by_project_dict_truncated.json"))
    
    config_dict["summary_report"] = {}
    config_dict["summary_report"]["columns"] = {"Col1":"<project_name>"}
    
    save_dir = save_and_send_reports_and_emails(authors_by_project_dict, pub_dict, config_dict, True)
    
    assert os.path.exists(os.path.join(save_dir, "summary_report.csv"))
    


def test_save_and_send_reports_and_emails_manual_name(config_dict, mocker):
    
    def emails(*args, **kwargs):
        return {"creation_date":"asdf", "emails":[]}
    mocker.patch("academic_tracker.athr_srch_modularized.athr_srch_emails_and_reports.create_project_reports_and_emails", emails)
    
    pub_dict = load_json(os.path.join("tests", "testing_files", "publication_dict_truncated.json"))
    authors_by_project_dict = load_json(os.path.join("tests", "testing_files", "authors_by_project_dict_truncated.json"))
    
    config_dict["summary_report"] = {}
    config_dict["summary_report"]["template"] = read_text_from_txt(os.path.join("tests", "testing_files", "athr_srch_build_loop_template_string.txt"))
    config_dict["summary_report"]["filename"] = "test_name.txt"
    
    save_dir = save_and_send_reports_and_emails(authors_by_project_dict, pub_dict, config_dict, True)
    
    assert os.path.exists(os.path.join(save_dir, "test_name.txt"))






@pytest.fixture(autouse=True)
def cleanup(request):
    """Cleanup a testing directory once we are finished."""
    def remove_test_dir():
        dir_contents = os.listdir()
        tracker_dirs = [folder for folder in dir_contents if re.match(r"tracker-(\d{10})", folder) or re.match(r"tracker-test-(\d{10})", folder)]
        for directory in tracker_dirs:
            shutil.rmtree(directory)
    request.addfinalizer(remove_test_dir)


# -*- coding: utf-8 -*-

import os

import pytest
import pickle
import requests
import xml.etree.ElementTree as ET
import json

import pymed

from fixtures import authors_dict
from academic_tracker.athr_srch_webio import search_PubMed_for_pubs, search_ORCID_for_pubs, search_Google_Scholar_for_pubs
from academic_tracker.athr_srch_webio import search_Crossref_for_pubs
from academic_tracker.fileio import load_json



@pytest.fixture(autouse=True)
def disable_network_calls(monkeypatch):
    def stunted_get():
        raise RuntimeError("Network access not allowed during testing!")
    monkeypatch.setattr(requests, "get", lambda *args, **kwargs: stunted_get())
    


@pytest.fixture
def config_dict_Hunter_only():
    return load_json(os.path.join("tests", "testing_files", "config_Hunter_only.json"))

@pytest.fixture
def original_queries():
    query_json = load_json(os.path.join("tests", "testing_files", "all_queries.json"))
    ## Convert PubMed dictionaries back to articles class.
    for author, pub_list in query_json["PubMed"].items():
        new_list = []
        for pub in pub_list:
            new_list.append(pymed.article.PubMedArticle(ET.fromstring(pub["xml"])))
        query_json["PubMed"][author] = new_list
    return query_json



## The expected_dicts here are generated in test_athr_srch_modularized.py.

def test_search_PubMed_for_pubs_first_pass(config_dict_Hunter_only, original_queries):
    expected_dict = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs1.json"))
    test_publication_dict, all_queries = search_PubMed_for_pubs({}, config_dict_Hunter_only["Authors"], "ptth222@uky.edu", original_queries["PubMed"])
    assert test_publication_dict == expected_dict
    assert all_queries == original_queries["PubMed"]


def test_search_PubMed_for_pubs_second_pass(config_dict_Hunter_only, original_queries):
    running_pubs = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs4.json"))
    expected_dict = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs5.json"))
    test_publication_dict, all_queries = search_PubMed_for_pubs(running_pubs, config_dict_Hunter_only["Authors"], "ptth222@uky.edu", original_queries["PubMed"])
    assert test_publication_dict == expected_dict
    assert all_queries == original_queries["PubMed"]
    

def test_search_PubMed_for_pubs_merge(config_dict_Hunter_only, original_queries):
    """PubMed is always ran first, so the merge code isn't triggered in other tests. 
    Feed a publication dict created from Google Scholar alone into prev_query to test the merge code."""
    running_pubs = load_json(os.path.join("tests", "testing_files", "solo_Google_Scholar.json"))
    expected_dict = load_json(os.path.join("tests", "testing_files", "PubMed_merge.json"))
    test_publication_dict, all_queries = search_PubMed_for_pubs(running_pubs, config_dict_Hunter_only["Authors"], "ptth222@uky.edu", original_queries["PubMed"])
    # with open(os.path.join("tests", "testing_files", "PubMed_merge_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(test_publication_dict, indent=2, sort_keys=True))
    assert test_publication_dict == expected_dict
    assert all_queries == original_queries["PubMed"]
    

def test_search_PubMed_for_pubs_continues(config_dict_Hunter_only, original_queries, mocker):
    "Feed in a prev_query that will trigger the continue lines."
    prev_queries = original_queries["PubMed"]
    prev_queries["Hunter Moseley"] = prev_queries["Hunter Moseley"][0:3]
    prev_queries["Hunter Moseley"][0] = prev_queries["Hunter Moseley"][0].toDict()
    prev_queries["Hunter Moseley"][1].publication_date = None
    mocker.patch("academic_tracker.athr_srch_webio.helper_functions.match_pub_authors_to_config_authors", 
                  side_effect=[[]])
    test_publication_dict, all_queries = search_PubMed_for_pubs({}, config_dict_Hunter_only["Authors"], "ptth222@uky.edu", prev_queries)
    assert test_publication_dict == {}
    ## If the type is not pymed.article.PubMedArticle, then it is not added to all_queries.
    del prev_queries["Hunter Moseley"][0]
    assert all_queries == prev_queries


def test_search_ORCID_for_pubs_first_pass(config_dict_Hunter_only, original_queries):
    expected_dict = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_PubMed", "running_pubs1.json"))
    test_publication_dict, all_queries = search_ORCID_for_pubs({}, "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
    assert test_publication_dict == expected_dict
    assert all_queries == original_queries["ORCID"]


def test_search_ORCID_for_pubs_second_pass(config_dict_Hunter_only, original_queries):
    running_pubs = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_PubMed", "running_pubs3.json"))
    expected_dict = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "no_PubMed", "running_pubs4.json"))
    test_publication_dict, all_queries = search_ORCID_for_pubs(running_pubs, "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
    assert test_publication_dict == expected_dict
    assert all_queries == original_queries["ORCID"]


def test_search_ORCID_for_pubs_merge(config_dict_Hunter_only, original_queries):
    "Previous tests did not trigger a merge because everything was merged in other calls on the first pass."
    running_pubs = load_json(os.path.join("tests", "testing_files", "solo_Google_Scholar.json"))
    expected_dict = load_json(os.path.join("tests", "testing_files", "ORCID_merge.json"))
    test_publication_dict, all_queries = search_ORCID_for_pubs(running_pubs, "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
    # with open(os.path.join("tests", "testing_files", "ORCID_merge_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(test_publication_dict, indent=2, sort_keys=True))
    assert test_publication_dict == expected_dict
    assert all_queries == original_queries["ORCID"]


def test_search_ORCID_for_pubs_misc(config_dict_Hunter_only, original_queries):
    "Manually modifying some ORCID results to trigger miscellaneous code not triggered by other tests."
    prev_queries = original_queries["ORCID"]
    prev_queries["Hunter Moseley"] = prev_queries["Hunter Moseley"][0:3]
    
    ##Change doi external ID to something else.
    prev_queries["Hunter Moseley"][0]["work-summary"][0]["external-ids"]["external-id"][0]["external-id-type"] = "asdf"
    prev_queries["Hunter Moseley"][0]["work-summary"][0]["external-ids"]["external-id"][0]["external-id-url"]["value"] = "asdf"
    prev_queries["Hunter Moseley"][1]["work-summary"][0]["external-ids"]["external-id"][0]["external-id-type"] = "pmid"
    prev_queries["Hunter Moseley"][1]["work-summary"][0]["external-ids"]["external-id"][0]["external-id-url"] = None
    del prev_queries["Hunter Moseley"][1]["work-summary"][1]
    ## Create an entry with no title.
    prev_queries["Hunter Moseley"][2]["work-summary"][0]["title"] = None
    
    ## Add an author with no ORCID to trigger continue.
    config_dict_Hunter_only["Authors"]["asdf"] = {}
    
    ## Add a collective_name to Hunter to create this author style instead of the usual one.
    config_dict_Hunter_only["Authors"]["Hunter Moseley"]["collective_name"] = "Some Collective"
    
    expected_dict = load_json(os.path.join("tests", "testing_files", "ORCID_misc.json"))
    test_publication_dict, all_queries = search_ORCID_for_pubs({}, "asdf", "asdf", config_dict_Hunter_only["Authors"], prev_queries)
    # with open(os.path.join("tests", "testing_files", "ORCID_misc_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(test_publication_dict, indent=2, sort_keys=True))
    assert test_publication_dict == expected_dict
    ## An empty "asdf" author gets added that wasn't there in the original run.
    del all_queries["asdf"]
    assert all_queries == prev_queries


def test_search_ORCID_for_pubs_no_IDs(config_dict_Hunter_only, original_queries, mocker, capsys):
    "If we can't find a DOI, pubmed id or title, test that a message is printed."
    prev_queries = original_queries["ORCID"]
    prev_queries["Hunter Moseley"] = prev_queries["Hunter Moseley"][0:1]
    prev_queries["Hunter Moseley"][0]["work-summary"][0]["external-ids"]["external-id"] = []
    
    mock_orcid = mocker.MagicMock()
    mock_orcid.configure_mock(
        **{
            "get_search_token_from_orcid.return_value": "asdf",
            "read_record_public.return_value": {"group":prev_queries["Hunter Moseley"]}
        }
    )
    
    mocker.patch("academic_tracker.athr_srch_webio.orcid.PublicAPI", 
                  side_effect=[mock_orcid])
    
    test_publication_dict, all_queries = search_ORCID_for_pubs({}, "asdf", "asdf", config_dict_Hunter_only["Authors"], None)
    captured = capsys.readouterr()
    # with open(os.path.join("tests", "testing_files", "ORCID_misc_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(test_publication_dict, indent=2, sort_keys=True))
    assert test_publication_dict == {}
    assert captured.out == ("Warning: Could not find a DOI, URL, or PMID for a "
                            "publication when searching ORCID. It will not be in the publications\n"
                            "Title: The metabolomics workbench file status website: a metadata "
                            "repository promoting FAIR principles of metabolomics data\n")
    assert all_queries == prev_queries


def test_search_Google_Scholar_for_pubs_first_pass(config_dict_Hunter_only, original_queries):
    expected_dict = load_json(os.path.join("tests", "testing_files", "solo_Google_Scholar.json"))
    test_publication_dict, all_queries = search_Google_Scholar_for_pubs({}, config_dict_Hunter_only["Authors"], "ptth222@uky.edu", original_queries["Google Scholar"])
    # with open(os.path.join("tests", "testing_files", "solo_Google_Scholar_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(test_publication_dict, indent=2, sort_keys=True))
    assert test_publication_dict == expected_dict
    assert all_queries == original_queries["Google Scholar"]


def test_search_Google_Scholar_for_pubs_second_pass(config_dict_Hunter_only, original_queries):
    running_pubs = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs6.json"))
    expected_dict = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs7.json"))
    test_publication_dict, all_queries = search_Google_Scholar_for_pubs(running_pubs, config_dict_Hunter_only["Authors"], "ptth222@uky.edu", original_queries["Google Scholar"])
    assert test_publication_dict == expected_dict
    assert all_queries == original_queries["Google Scholar"]
    

def test_search_Google_Scholar_for_pubs_merge(config_dict_Hunter_only, original_queries):
    "Previous tests did not trigger a merge because everything was merged in other calls on the first pass."
    running_pubs = load_json(os.path.join("tests", "testing_files", "solo_Crossref.json"))
    expected_dict = load_json(os.path.join("tests", "testing_files", "Google_Scholar_merge.json"))
    test_publication_dict, all_queries = search_Google_Scholar_for_pubs(running_pubs, config_dict_Hunter_only["Authors"], "ptth222@uky.edu", original_queries["Google Scholar"])
    # with open(os.path.join("tests", "testing_files", "Google_Scholar_merge_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(test_publication_dict, indent=2, sort_keys=True))
    assert test_publication_dict == expected_dict
    assert all_queries == original_queries["Google Scholar"]


def test_search_Google_Scholar_for_pubs_misc(config_dict_Hunter_only, original_queries):
    "Manually modifying some Google Scholar results to trigger miscellaneous code not triggered by other tests."
    prev_queries = original_queries["Google Scholar"]
    prev_queries["Hunter Moseley"] = prev_queries["Hunter Moseley"][0:1]
    prev_queries["Hunter Moseley"][0]["bib"]["pub_year"] = 2022
    
    ## Add an author with no scholar_id to trigger continue.
    config_dict_Hunter_only["Authors"]["asdf"] = {}
    
    ## Add a collective_name to Hunter to create this author style instead of the usual one.
    config_dict_Hunter_only["Authors"]["Hunter Moseley"]["collective_name"] = "Some Collective"
    
    expected_dict = load_json(os.path.join("tests", "testing_files", "Google_Scholar_misc.json"))
    test_publication_dict, all_queries = search_Google_Scholar_for_pubs({}, config_dict_Hunter_only["Authors"], "asdf", prev_queries)
    # with open(os.path.join("tests", "testing_files", "Google_Scholar_misc_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(test_publication_dict, indent=2, sort_keys=True))
    assert test_publication_dict == expected_dict
    ## An empty "asdf" author gets added that wasn't there in the original run.
    del all_queries["asdf"]
    assert all_queries == prev_queries


def test_search_Google_Scholar_for_pubs_query_error(config_dict_Hunter_only, original_queries, capsys, mocker):
    "Mock an error to make sure it is caught and a message is printed."    
    mocker.patch("academic_tracker.athr_srch_webio.scholarly.scholarly.search_author_id", 
                  side_effect=[Exception])
    test_publication_dict, all_queries = search_Google_Scholar_for_pubs({}, config_dict_Hunter_only["Authors"], "asdf", None)
    captured = capsys.readouterr()
    
    assert ("Warning: The \"scholar_id\" for author Hunter Moseley is probably "
            "incorrect, an error occured when trying to query Google Scholar.\n") in captured.out
    assert r'Traceback (most recent call last):' in captured.out
    assert r'Exception' in captured.out


def test_search_Google_Scholar_for_pubs_scholar_id_mismatch(config_dict_Hunter_only, original_queries, mocker):
    "Make sure that if scholar_ids don't match the author is skipped."    
    mocker.patch("academic_tracker.athr_srch_webio.scholarly.scholarly.search_author_id", 
                  side_effect=[{"scholar_id": "qwer"}])
    test_publication_dict, all_queries = search_Google_Scholar_for_pubs({}, config_dict_Hunter_only["Authors"], "asdf", None)
    assert test_publication_dict == {}
    assert all_queries == {'Hunter Moseley': []}
    

def test_search_Google_Scholar_for_pubs_fill_author(config_dict_Hunter_only, original_queries, mocker):
    "Mock it so that the fill function is called."
    publications = original_queries["Google Scholar"]["Hunter Moseley"][0:1]
    publications[0]["bib"]["pub_year"] = 2022
    
    mocker.patch("academic_tracker.athr_srch_webio.scholarly.scholarly.search_author_id", 
                  side_effect=[{"scholar_id": config_dict_Hunter_only["Authors"]["Hunter Moseley"]["scholar_id"], "publications":{}}])
    mocker.patch("academic_tracker.athr_srch_webio.scholarly.scholarly.fill", 
                  side_effect=[{"publications": publications}])
    mocker.patch("academic_tracker.athr_srch_webio.webio.get_DOI_from_Crossref", 
                  side_effect=[publications[0]["doi"]])
    
    expected_dict = load_json(os.path.join("tests", "testing_files", "Google_Scholar_misc.json"))
    expected_dict["https://doi.org/10.1016/s0959-440x(99)00019-6"]["authors"] = [{
                                                                                    "ORCID": "0000-0003-3995-5368",
                                                                                    "affiliation": "kentucky",
                                                                                    "author_id": "Hunter Moseley",
                                                                                    "firstname": "Hunter",
                                                                                    "initials": None,
                                                                                    "lastname": "Moseley"
                                                                                  }]
    test_publication_dict, all_queries = search_Google_Scholar_for_pubs({}, config_dict_Hunter_only["Authors"], "asdf", None)
    assert test_publication_dict == expected_dict
    assert all_queries == {'Hunter Moseley': publications}


def test_search_Google_Scholar_for_pubs_fill_pub(config_dict_Hunter_only, original_queries, mocker):
    "Test that fill publication works."
    publications = original_queries["Google Scholar"]["Hunter Moseley"][0:1]
    publications[0]["bib"]["pub_year"] = 2022
    publications[0]["pub_url"] = "asdf"
    
    mocker.patch("academic_tracker.athr_srch_webio.scholarly.scholarly.search_author_id", 
                  side_effect=[{"scholar_id": config_dict_Hunter_only["Authors"]["Hunter Moseley"]["scholar_id"], "publications":{}}])
    mocker.patch("academic_tracker.athr_srch_webio.scholarly.scholarly.fill", 
                  side_effect=[{"publications": publications}, publications[0]])
    mocker.patch("academic_tracker.athr_srch_webio.webio.get_DOI_from_Crossref", 
                  side_effect=[None])
    
    expected_dict = load_json(os.path.join("tests", "testing_files", "Google_Scholar_misc.json"))
    expected_dict["https://doi.org/10.1016/s0959-440x(99)00019-6"]["authors"] = [{
                                                                                    "ORCID": "0000-0003-3995-5368",
                                                                                    "affiliation": "kentucky",
                                                                                    "author_id": "Hunter Moseley",
                                                                                    "firstname": "Hunter",
                                                                                    "initials": None,
                                                                                    "lastname": "Moseley"
                                                                                  }]
    expected_dict["asdf"] = expected_dict["https://doi.org/10.1016/s0959-440x(99)00019-6"]
    del expected_dict["https://doi.org/10.1016/s0959-440x(99)00019-6"]
    expected_dict["asdf"]["doi"] = None
    
    test_publication_dict, all_queries = search_Google_Scholar_for_pubs({}, config_dict_Hunter_only["Authors"], "asdf", None)
    assert test_publication_dict == expected_dict
    assert all_queries == {'Hunter Moseley': publications}


def test_search_Google_Scholar_for_pubs_no_pub_ID(config_dict_Hunter_only, original_queries, capsys, mocker):
    "Test that a message is printed when a pub ID cannot be found."
    publications = original_queries["Google Scholar"]["Hunter Moseley"][0:1]
    publications[0]["bib"]["pub_year"] = 2022
    
    mocker.patch("academic_tracker.athr_srch_webio.scholarly.scholarly.search_author_id", 
                  side_effect=[{"scholar_id": config_dict_Hunter_only["Authors"]["Hunter Moseley"]["scholar_id"], "publications":{}}])
    mocker.patch("academic_tracker.athr_srch_webio.scholarly.scholarly.fill", 
                  side_effect=[{"publications": publications}, publications[0]])
    mocker.patch("academic_tracker.athr_srch_webio.webio.get_DOI_from_Crossref", 
                  side_effect=[None])
        
    test_publication_dict, all_queries = search_Google_Scholar_for_pubs({}, config_dict_Hunter_only["Authors"], "asdf", None)
    captured = capsys.readouterr()
    
    assert test_publication_dict == {}
    assert all_queries == {'Hunter Moseley': publications}
    assert captured.out == ("Warning: Could not find a DOI, URL, or PMID for a publication "
                            "when searching Google Scholar. It will not be in the publications.\n"
                            "Title: Automated analysis of NMR assignments and structures for proteins\n")


def test_search_Crossref_for_pubs_first_pass(config_dict_Hunter_only, original_queries):
    expected_dict = load_json(os.path.join("tests", "testing_files", "solo_Crossref.json"))
    test_publication_dict, all_queries = search_Crossref_for_pubs({}, config_dict_Hunter_only["Authors"], "ptth222@uky.edu", original_queries["Crossref"])
    # with open(os.path.join("tests", "testing_files", "solo_Crossref_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(test_publication_dict, indent=2, sort_keys=True))
    assert test_publication_dict == expected_dict
    assert all_queries == original_queries["Crossref"]


def test_search_Crossref_for_pubs_second_pass(config_dict_Hunter_only, original_queries):
    running_pubs = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs7.json"))
    expected_dict = load_json(os.path.join("tests", "testing_files", "intermediate_results", "author_search", "all", "running_pubs8.json"))
    test_publication_dict, all_queries = search_Crossref_for_pubs(running_pubs, config_dict_Hunter_only["Authors"], "ptth222@uky.edu", original_queries["Crossref"])
    assert test_publication_dict == expected_dict
    assert all_queries == original_queries["Crossref"]


def test_search_Crossref_for_pubs_merge(config_dict_Hunter_only, original_queries):
    "Previous tests did not trigger a merge because everything was merged in other calls on the first pass."
    running_pubs = load_json(os.path.join("tests", "testing_files", "solo_Google_Scholar.json"))
    expected_dict = load_json(os.path.join("tests", "testing_files", "Crossref_merge.json"))
    test_publication_dict, all_queries = search_Crossref_for_pubs(running_pubs, config_dict_Hunter_only["Authors"], "ptth222@uky.edu", original_queries["Crossref"])
    # with open(os.path.join("tests", "testing_files", "Crossref_merge_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(test_publication_dict, indent=2, sort_keys=True))
    assert test_publication_dict == expected_dict
    assert all_queries == original_queries["Crossref"]


def test_search_Crossref_for_pubs_misc(config_dict_Hunter_only, original_queries):
    "Manually modifying some Crossref results to trigger miscellaneous code not triggered by other tests."
    prev_queries = original_queries["Crossref"]
    prev_queries["Hunter Moseley"] = prev_queries["Hunter Moseley"][0:4]
    ## get rid of sources of dates so values will be None and continue will be triggered.
    del prev_queries["Hunter Moseley"][0]["published"]
    del prev_queries["Hunter Moseley"][0]["published-online"]
    ## Make publication year less than cutoff to trigger continue.
    prev_queries["Hunter Moseley"][1]["published"]["date-parts"][0][0] = 1999
    ## delete URL and DOI to trigger continue.
    del prev_queries["Hunter Moseley"][2]["DOI"]
    del prev_queries["Hunter Moseley"][2]["URL"]
    del prev_queries["Hunter Moseley"][2]["link"]
            
    expected_dict = load_json(os.path.join("tests", "testing_files", "Crossref_misc.json"))
    test_publication_dict, all_queries = search_Crossref_for_pubs({}, config_dict_Hunter_only["Authors"], "asdf", prev_queries)
    # with open(os.path.join("tests", "testing_files", "Crossref_misc_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(test_publication_dict, indent=2, sort_keys=True))
    assert test_publication_dict == expected_dict
    assert all_queries == prev_queries


def test_search_Crossref_for_pubs_no_pub_ID(config_dict_Hunter_only, original_queries, capsys, mocker):
    "Test that a message is printed when a pub ID cannot be found."
    prev_queries = original_queries["Crossref"]
    prev_queries["Hunter Moseley"] = prev_queries["Hunter Moseley"][0:1]
    del prev_queries["Hunter Moseley"][0]["DOI"]
    del prev_queries["Hunter Moseley"][0]["URL"]
    del prev_queries["Hunter Moseley"][0]["link"]
    
    mock_crossref = mocker.MagicMock()
    mock_crossref.configure_mock(
        **{
            "works.return_value": {"message":{"items":prev_queries["Hunter Moseley"]}}
        }
    )
    
    mocker.patch("academic_tracker.athr_srch_webio.habanero.Crossref", 
                  side_effect=[mock_crossref])
            
    test_publication_dict, all_queries = search_Crossref_for_pubs({}, config_dict_Hunter_only["Authors"], "asdf", None)
    captured = capsys.readouterr()
    
    assert test_publication_dict == {}
    assert all_queries == prev_queries
    assert captured.out == ("Warning: Could not find a DOI or external URL for a publication when "
                            "searching Crossref. It will not be in the publications.\n"
                            "Title: kegg_pull: a software package for the RESTful access and "
                            "pulling from the Kyoto Encyclopedia of Gene and Genomes\n")









# -*- coding: utf-8 -*-
"""
Author Search Webio
~~~~~~~~~~~~~~~~~~~

Internet interfacing for author_search.
"""


import time
import copy
import traceback

import pymed
import orcid
import scholarly
import habanero

from . import helper_functions
from . import webio



TOOL = webio.TOOL
DOI_URL = webio.DOI_URL

PUBLICATION_TEMPLATE = webio.PUBLICATION_TEMPLATE


## TODO get with pymed and add grants and pmcid to PubMedArticle class.
def search_PubMed_for_pubs(running_pubs, authors_json, from_email, prev_query=None):
    """Searhes PubMed for publications by each author.
    
    For each author in authors_json PubMed is queried for the publications. The list of publications is then filtered 
    by affiliations and cutoff_year. If the publication is in the of running_pubs then it tries to fill in missing 
    information from this source. If the author doesn't have at least one matching affiliation then the publication 
    is skipped. If the publication was published before the cutoff_year then it is skipped. If prev_query is given, then 
    publications will be taken from it instead of querying PubMed again. 
    
    Args:
        running_pubs (dict): dictionary of publications matching the JSON schema for publications.
        authors_json (dict): keys are authors and values are author attributes. Matches Authors section of configuration JSON schema.
        from_email (str): used in the query to PubMed.
        prev_query (dict|None): a dictionary containing publications from a previous call to this function. {author1: [pub1, ...], ...}
        
    Returns:
        running_pubs (dict): keys are publication ids and values are a dictionary with publication attributes
        all_pubs (dict): a dictionary where the keys are the authors in authors_json and the values are a list of the publications queried for them.
    """
    
    ## Some helpful code to get the xml back as text. import xml.etree.ElementTree as ET    ET.tostring(Element)    ET.ElementTree(element).write('path')
    # initiate PubMed API
    pubmed = pymed.PubMed(tool=TOOL, email=from_email)
    
    all_pubs = {}
    
    ########################
    # loop through list of authors and request a list of their publications
    ########################
    for author, author_attributes in authors_json.items():
        all_pubs[author] = []
        
        publications = pubmed.query(author_attributes["pubmed_name_search"], max_results=500) if not prev_query else prev_query[author]
        
        ## Unpacking pub from publications appears to be the slowest part of the code.
        ## publications is an iterator that is broken up into batches and there are noticeable slow downs each time a new batch is fetched.
        for pub in publications:
            if not isinstance(pub, pymed.article.PubMedArticle):
                continue
            all_pubs[author].append(pub)
            
            pub_id, pub_dict = helper_functions.create_pub_dict_for_saving_PubMed(pub)
            
            if matching_pub_id := helper_functions.get_pub_id_in_publication_dict(pub_id, pub.title, running_pubs):
                if "PubMed" in running_pubs[matching_pub_id]["queried_sources"]:
                    continue
                
                helper_functions._merge_pub_dicts(running_pubs[matching_pub_id], pub_dict)
                running_pubs[matching_pub_id]["queried_sources"].append("PubMed")
            else:
                    
                ## Sometimes the publication_date can be None, so just skip it.
                if not pub.publication_date:
                    continue
                publication_date = int(str(pub.publication_date)[:4])
                
                ## if the publication date is before the cutoff year then skip.
                if publication_date < author_attributes["cutoff_year"]:
                    continue
                
                author_list = helper_functions.match_pub_authors_to_config_authors(authors_json, pub_dict["authors"])
        
                ## If no authors were matched then go to the next publication. Note that this is not uncommon because PubMed returns publications for authors who were just colloborators.
                if not author_list:
                    continue
                    
                pub_dict["authors"] = author_list
                pub_dict["queried_sources"] = ["PubMed"]
                running_pubs[pub_id] = pub_dict
                    
            
        # don't piss off NCBI
        time.sleep(1)
        
    return running_pubs, all_pubs


       
        
        
def search_ORCID_for_pubs(running_pubs, ORCID_key, ORCID_secret, authors_json, prev_query=None):
    """Searhes ORCID for publications by each author.
    
    For each author in authors_json ORCID is queried for the publications. The list of publications is then filtered 
    by affiliations and cutoff_year. If the author doesn't have at least one matching affiliation, then the publication 
    is skipped. If the publication was published before the cutoff_year, then it is skipped. If prev_query is given, then publications 
    will be taken from it instead of querying ORCID again. 
    
    Args:
        running_pubs (dict): dictionary of publications matching the JSON schema for publications.
        ORCID_key (str): string of the app key ORCID gives when you register the app with them
        ORCID_secret (str): string of the secret ORCID gives when you register the app with them
        authors_json (dict): keys are authors and values are author attributes. Matches authors JSON schema.
        prev_query (dict|None): a dictionary containing publications from a previous call to this function. {author1: [pub1, ...], ...}
        
    Returns:
        running_pubs (dict): keys are publication ids and values are a dictionary with publication attributes
        all_pubs (dict): a dictionary where the keys are the authors in authors_json and the values are a list of the publications queried for them.
    """
    
    if prev_query is None:
        api = orcid.PublicAPI(ORCID_key, ORCID_secret)
        search_token = api.get_search_token_from_orcid()
    
    all_pubs = {}

    for author, authors_attributes in authors_json.items():
        all_pubs[author] = []
        
        if not "ORCID" in authors_attributes:
            continue
        
        works = api.read_record_public(authors_attributes["ORCID"], 'works', search_token)["group"] if prev_query is None else prev_query[author]
        
        for work in works:
            all_pubs[author].append(work)
            
            title = None
            doi = None
            external_url = None
            publication_year = None
            publication_month = None
            publication_day = None
            pmid = None
            ## If the work is not a journal article then skip it.
            work_is_a_journal_article = True
            work_before_relevant_year = False
            for work_summary in work["work-summary"]:
                
                if work_summary["type"] != "JOURNAL_ARTICLE":
                    work_is_a_journal_article = False
                
                
                if work_summary["publication-date"]:
                    if not publication_year and work_summary["publication-date"]["year"]:
                        publication_year = int(work_summary["publication-date"]["year"]["value"])
                        
                    if not publication_month and work_summary["publication-date"]["month"]:
                        publication_month = int(work_summary["publication-date"]["month"]["value"])
                        
                    if not publication_day and work_summary["publication-date"]["day"]:
                        publication_day = int(work_summary["publication-date"]["day"]["value"])

                    if publication_year is not None and publication_year < authors_attributes["cutoff_year"]:
                        work_before_relevant_year = True
                
                
                if work_summary["title"] and not title:
                    title = work_summary["title"]["title"]["value"]
                
                if not doi:
                    for external_id in work_summary["external-ids"]["external-id"]:
                        if external_id["external-id-type"] == "doi":
                            doi = external_id["external-id-value"].lower()
                        elif external_id["external-id-url"]:
                            external_url = external_id["external-id-url"]["value"]
                        elif external_id["external-id-type"] == "pmid":
                            pmid = external_id["external-id-value"]
                
                if title and doi and publication_year and publication_month and publication_day:
                    break
            
                        
            ## Try to find a way to give the publication an ID.
            if doi:
                pub_id = DOI_URL + doi
            elif external_url:
                pub_id = external_url
            elif pmid:
                pub_id = pmid
            elif not prev_query and title is not None:
                helper_functions.vprint("Warning: Could not find a DOI, URL, or PMID for a publication when searching ORCID. It will not be in the publications", verbosity=1)
                helper_functions.vprint("Title: " + title, verbosity=1)
                continue
            else:
                continue
            
            ## Pretty sure the title is never None, but if it is then there will be an error, so skip.
            if title is None:
                continue
            
            ## Pull out relevant information from ORCID.
            pub_dict = copy.deepcopy(PUBLICATION_TEMPLATE)
            if doi:
                pub_dict["doi"] = doi
            if title:
                pub_dict["title"] = title
            if publication_year:
                pub_dict["publication_date"]["year"] = publication_year
            if publication_month:
                pub_dict["publication_date"]["month"] = publication_month
            if publication_day:
                pub_dict["publication_date"]["day"] = publication_day
            if pmid:
                pub_dict["pubmed_id"] = pmid
                
           
            authors_dict = {}
            authors_dict["ORCID"] = authors_attributes.get("ORCID")
            authors_dict["author_id"] = author
            if "collective_name" in authors_attributes:
                authors_dict["collectivename"] = authors_attributes["collective_name"]
            else:
                authors_dict["affiliation"] = "\n".join(authors_attributes["affiliations"]) if authors_attributes["affiliations"] else None
                authors_dict["firstname"] = authors_attributes["first_name"]
                authors_dict["initials"] = None
                authors_dict["lastname"] = authors_attributes["last_name"]
            
            pub_dict["authors"] = [authors_dict]
           
            
            ## If the publication is already in running_pubs then try to update missing information.
            if matching_pub_id := helper_functions.get_pub_id_in_publication_dict(pub_id, title, running_pubs):
                if "ORCID" in running_pubs[matching_pub_id]["queried_sources"]:
                    continue
                
                helper_functions._merge_pub_dicts(running_pubs[matching_pub_id], pub_dict)
                running_pubs[matching_pub_id]["queried_sources"].append("ORCID")
            
            else:
            
                if work_before_relevant_year or not work_is_a_journal_article or not publication_year:
                    continue
                    
                pub_dict["queried_sources"] = ["ORCID"]
                running_pubs[pub_id] = pub_dict
                
        time.sleep(1)
        
    return running_pubs, all_pubs




def search_Google_Scholar_for_pubs(running_pubs, authors_json, mailto_email, prev_query=None):
    """Searhes Google Scholar for publications by each author.
    
    For each author in authors_json Google Scholar is queried for the publications. The list of publications is then filtered 
    by affiliations and cutoff_year. If the author doesn't have at least one matching affiliation, then the publication is 
    skipped. If the publication was published before the cutoff_year, then it is skipped. If prev_query is given, then publications 
    will be taken from it instead of querying Google Scholar again. 
    
    Args:
        running_pubs (dict): dictionary of publications matching the JSON schema for publications.
        authors_json (dict): keys are authors and values are author attributes. Matches authors JSON schema.
        mailto_email (str): used in the query to Crossref when trying to find DOIs for the articles.
        prev_query (dict|None): a dictionary containing publications from a previous call to this function. {author1: [pub1, ...], ...}
        
    Returns:
        running_pubs (dict): keys are pulication ids and values are a dictionary with publication attributes
        all_pubs (dict): a dictionary where the keys are the authors in authors_json and the values are a list of the publications queried for them.
    """
    all_pubs = {}
    for author, authors_attributes in authors_json.items():
        all_pubs[author] = []
        
        if not "scholar_id" in authors_attributes:
            continue
        
        ## Either query Google Scholar or use the prev_query parameter.
        if not prev_query:
            try:
                queried_author = scholarly.scholarly.search_author_id(authors_attributes["scholar_id"])
            except:
                message = "Warning: The \"scholar_id\" for author " + author + " is probably incorrect, an error occured when trying to query Google Scholar.\n"
                message += traceback.format_exc()
                helper_functions.vprint(message, verbosity=1)    
                continue
            
            if not queried_author["scholar_id"] == authors_attributes["scholar_id"]:
                continue
            
            ## Note that fill modifies the passed dictionary directly, but this is easier to mock in unit tests.
            queried_author = scholarly.scholarly.fill(queried_author, sections=["publications"])
            publications = queried_author["publications"]
        else:
            publications = prev_query[author]
        
        ## Loop over queried publications.
        for i, pub in enumerate(publications):
            all_pubs[author].append(pub)
            
            ## Determine the pub_id
            title = pub["bib"]["title"]
            doi = webio.get_DOI_from_Crossref(title, mailto_email) if prev_query is None else pub["doi"]
            all_pubs[author][i]["doi"] = doi
            if doi:
                pub_id = DOI_URL + doi
            else:
                if prev_query is None:
                    pub = scholarly.scholarly.fill(pub)
                ## The fill method modifies the original pub I think, so this line isn't necessary.
                # all_pubs[author][-1] = pub
                if "pub_url" in pub:
                    pub_id = pub["pub_url"]
                elif not prev_query:
                    helper_functions.vprint("Warning: Could not find a DOI, URL, or PMID for a publication when searching Google Scholar. It will not be in the publications.", verbosity=1)
                    helper_functions.vprint("Title: " + title, verbosity=1)
                    continue
                else:
                    continue
            
            ## Build pub_dict
            publication_year = int(pub["bib"]["pub_year"]) if "pub_year" in pub["bib"] else None
            
            pub_dict = copy.deepcopy(PUBLICATION_TEMPLATE)
            if doi:
                pub_dict["doi"] = doi
            if title:
                pub_dict["title"] = title
            if publication_year:
                pub_dict["publication_date"]["year"] = publication_year
            
            authors_dict = {}
            authors_dict["ORCID"] = authors_attributes.get("ORCID")
            authors_dict["author_id"] = author
            if "collective_name" in authors_attributes:
                authors_dict["collectivename"] = authors_attributes["collective_name"]
            else:
                authors_dict["affiliation"] = "\n".join(authors_attributes["affiliations"]) if authors_attributes["affiliations"] else None
                authors_dict["firstname"] = authors_attributes["first_name"]
                authors_dict["initials"] = None
                authors_dict["lastname"] = authors_attributes["last_name"]
            
            pub_dict["authors"] = [authors_dict]
            
            
            ## If the publication is already in running_pubs then try to update missing information.
            if matching_pub_id := helper_functions.get_pub_id_in_publication_dict(pub_id, title, running_pubs):
                if "Google Scholar" in running_pubs[matching_pub_id]["queried_sources"]:
                    continue
                
                helper_functions._merge_pub_dicts(running_pubs[matching_pub_id], pub_dict)
                running_pubs[matching_pub_id]["queried_sources"].append("Google Scholar")
            
            else:
            
                ## Check if the publication year is in range.
                if not publication_year or publication_year < authors_attributes["cutoff_year"]:
                    continue
                
                pub_dict["queried_sources"] = ["Google Scholar"]
                running_pubs[pub_id] = pub_dict
                            
        time.sleep(1)
            
    return running_pubs, all_pubs





def search_Crossref_for_pubs(running_pubs, authors_json, mailto_email, prev_query=None):
    """Searhes Crossref for publications by each author.
    
    For each author in authors_json Crossref is queried for the publications. The list of publications is then filtered 
    by affiliations and cutoff_year. If the author doesn't have at least one matching affiliation, then the publication 
    is skipped. If the publication was published before the cutoff_year, then it is skipped. Each publication is then 
    determined to have citations for any of the grants in the author's grants. If prev_query is given, then publications 
    will be taken from it instead of querying Crossref again. 
    
    Args:
        running_pubs (dict): dictionary of publications matching the JSON schema for publications.
        authors_json (dict): keys are authors and values are author attributes. Matches authors JSON schema.
        mailto_email (str): used in the query to Crossref.
        prev_query (dict|None): a dictionary containing publications from a previous call to this function. {author1: [pub1, ...], ...}
        
    Returns:
        running_pubs (dict): keys are pulication ids and values are a dictionary with publication attributes
        all_pubs (dict): a dictionary where the keys are the authors in authors_json and the values are a list of the publications queried for them.
    """
    
    cr = habanero.Crossref(ua_string = "Academic Tracker (mailto:" + mailto_email + ")")
    
    all_pubs = {}
    for author, authors_attributes in authors_json.items():
        all_pubs[author] = []
        
        ## Query Crossref or use prev_query.
        if not prev_query:
            results = cr.works(query_author = authors_attributes["pubmed_name_search"], 
                               filter = {"type":"journal-article", "from-pub-date":str(authors_attributes["cutoff_year"])}, 
                               limit = 300)
            publications = results["message"]["items"]
        else:
            publications = prev_query[author]
        
        ## Loop over publications.
        for work in publications:
            all_pubs[author].append(work)
            
            pub_id, pub_dict = helper_functions.create_pub_dict_for_saving_Crossref(work, prev_query)
            
            if pub_id is None:
                continue
            
            
            ## If the publication is already in running_pubs then try to update missing information.
            if matching_pub_id := helper_functions.get_pub_id_in_publication_dict(pub_id, pub_dict["title"], running_pubs):
                if "Crossref" in running_pubs[matching_pub_id]["queried_sources"]:
                    continue
                
                helper_functions._merge_pub_dicts(running_pubs[matching_pub_id], pub_dict)
                running_pubs[matching_pub_id]["queried_sources"].append("Crossref")
            
            else:
            
                if not pub_dict["publication_date"]["year"] or pub_dict["publication_date"]["year"] < authors_attributes["cutoff_year"]:
                    continue
                
                author_list = helper_functions.match_pub_authors_to_config_authors(authors_json, pub_dict["authors"])
                ## If the author_list is empty then there were no matching authors, continue.
                if not author_list:
                    continue
            
                pub_dict["authors"] = author_list
                pub_dict["queried_sources"] = ["Crossref"]
                running_pubs[pub_id] = pub_dict
            
        time.sleep(1)
            
            
    return running_pubs, all_pubs



"""

This package has the following modules:
    
``user_input_checking``
    This module contains functions for checking user input.

``tracker_schema``
    This module contains the schema for validating user input.
    
``athr_srch_modularized``
    This module contains functions to complete the author_search command modularized into pieces.
    
``athr_srch_webio``
    This module contains functions for author_search to interface with the internet.
    
``athr_srch_emails_and_reports``
    This module contains functions to create emails and reports for author_search.
    
``ref_srch_modularized``
    This module contains functions to complete the reference_search command modularized into pieces.
    
``ref_srch_webio``
    This module contains functions for reference_search to interface with the internet.
    
``ref_srch_emails_and_reports``
    This module contains functions to create emails and reports for reference_search.
    
``citation_parsing``
    This module contains functions for parsing references and citations for reference_search.

``fileio``
    This module contains functions for reading and writing files.
    
``helper_functions``
    This module contains functions that help the other modules function. The functions do things such as fuzzy matching, regex searching, and printing.

``emails_and_reports_helpers``
    This module contains functions that help create emails and reports.
        
``webio``
    This module contains general functions for interfacing with the internet.

"""
try:
    # -- Distribution mode --
    # import from _version.py generated by setuptools_scm during release
    from ._version import version as __version__
except ImportError:
    # -- Source mode --
    # use setuptools_scm to get the current version from src using git
    from setuptools_scm import get_version as _gv
    from os import path as _path
    __version__ = _gv(_path.join(_path.dirname(__file__), _path.pardir))



# -*- coding: utf-8 -*-
"""
Regenerate intermediate testing files, must be ran from academic_tracker directory.
"""

import json
import copy
import os

import pymed
import xml.etree.ElementTree as ET

from academic_tracker.fileio import load_json
from academic_tracker.athr_srch_webio import search_PubMed_for_pubs, search_ORCID_for_pubs, search_Google_Scholar_for_pubs, search_Crossref_for_pubs



original_queries = load_json(os.path.join("tests", "testing_files", "all_queries.json"))
## Convert PubMed dictionaries back to articles class.
for author, pub_list in original_queries["PubMed"].items():
    new_list = []
    for pub in pub_list:
        new_list.append(pymed.article.PubMedArticle(ET.fromstring(pub["xml"])))
    original_queries["PubMed"][author] = new_list


config_dict_Hunter_only = load_json(os.path.join("tests", "testing_files", "config_Hunter_only.json"))




running_pubs = {}
running_pubs1, _ = search_PubMed_for_pubs(running_pubs, config_dict_Hunter_only["Authors"], "asdf", original_queries["PubMed"])
running_pubs2, _ = search_ORCID_for_pubs(copy.deepcopy(running_pubs1), "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
running_pubs3, _ = search_Google_Scholar_for_pubs(copy.deepcopy(running_pubs2), config_dict_Hunter_only["Authors"], "asdf", original_queries["Google Scholar"])
running_pubs4, _ = search_Crossref_for_pubs(copy.deepcopy(running_pubs3), config_dict_Hunter_only["Authors"], "asdf", original_queries["Crossref"])

running_pubs5, _ = search_PubMed_for_pubs(copy.deepcopy(running_pubs4), config_dict_Hunter_only["Authors"], "asdf", original_queries["PubMed"])
running_pubs6, _ = search_ORCID_for_pubs(copy.deepcopy(running_pubs5), "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
running_pubs7, _ = search_Google_Scholar_for_pubs(copy.deepcopy(running_pubs6), config_dict_Hunter_only["Authors"], "asdf", original_queries["Google Scholar"])
running_pubs8, _ = search_Crossref_for_pubs(copy.deepcopy(running_pubs7), config_dict_Hunter_only["Authors"], "asdf", original_queries["Crossref"])

with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "all", "running_pubs1.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs1, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "all", "running_pubs2.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs2, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "all", "running_pubs3.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs3, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "all", "running_pubs4.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs4, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "all", "running_pubs5.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs5, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "all", "running_pubs6.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs6, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "all", "running_pubs7.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs7, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "all", "running_pubs8.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs8, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "all", "publication_dict.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs8, indent=2, sort_keys=True))



running_pubs = {}
running_pubs1, _ = search_PubMed_for_pubs(running_pubs, config_dict_Hunter_only["Authors"], "asdf", original_queries["PubMed"])
running_pubs2, _ = search_Google_Scholar_for_pubs(copy.deepcopy(running_pubs1), config_dict_Hunter_only["Authors"], "asdf", original_queries["Google Scholar"])
running_pubs3, _ = search_Crossref_for_pubs(copy.deepcopy(running_pubs2), config_dict_Hunter_only["Authors"], "asdf", original_queries["Crossref"])

running_pubs4, _ = search_PubMed_for_pubs(copy.deepcopy(running_pubs3), config_dict_Hunter_only["Authors"], "asdf", original_queries["PubMed"])
running_pubs5, _ = search_Google_Scholar_for_pubs(copy.deepcopy(running_pubs4), config_dict_Hunter_only["Authors"], "asdf", original_queries["Google Scholar"])
running_pubs6, _ = search_Crossref_for_pubs(copy.deepcopy(running_pubs5), config_dict_Hunter_only["Authors"], "asdf", original_queries["Crossref"])

with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_ORCID", "running_pubs1.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs1, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_ORCID", "running_pubs2.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs2, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_ORCID", "running_pubs3.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs3, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_ORCID", "running_pubs4.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs4, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_ORCID", "running_pubs5.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs5, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_ORCID", "running_pubs6.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs6, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_ORCID", "publication_dict.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs6, indent=2, sort_keys=True))



running_pubs = {}
running_pubs1, _ = search_PubMed_for_pubs(running_pubs, config_dict_Hunter_only["Authors"], "asdf", original_queries["PubMed"])
running_pubs2, _ = search_ORCID_for_pubs(copy.deepcopy(running_pubs1), "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
running_pubs3, _ = search_Google_Scholar_for_pubs(copy.deepcopy(running_pubs2), config_dict_Hunter_only["Authors"], "asdf", original_queries["Google Scholar"])

running_pubs4, _ = search_PubMed_for_pubs(copy.deepcopy(running_pubs3), config_dict_Hunter_only["Authors"], "asdf", original_queries["PubMed"])
running_pubs5, _ = search_ORCID_for_pubs(copy.deepcopy(running_pubs4), "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
running_pubs6, _ = search_Google_Scholar_for_pubs(copy.deepcopy(running_pubs5), config_dict_Hunter_only["Authors"], "asdf", original_queries["Google Scholar"])

with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_Crossref", "running_pubs1.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs1, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_Crossref", "running_pubs2.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs2, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_Crossref", "running_pubs3.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs3, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_Crossref", "running_pubs4.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs4, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_Crossref", "running_pubs5.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs5, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_Crossref", "running_pubs6.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs6, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_Crossref", "publication_dict.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs6, indent=2, sort_keys=True))



running_pubs = {}
running_pubs1, _ = search_PubMed_for_pubs(running_pubs, config_dict_Hunter_only["Authors"], "asdf", original_queries["PubMed"])
running_pubs2, _ = search_ORCID_for_pubs(copy.deepcopy(running_pubs1), "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
running_pubs3, _ = search_Crossref_for_pubs(copy.deepcopy(running_pubs2), config_dict_Hunter_only["Authors"], "asdf", original_queries["Crossref"])

running_pubs4, _ = search_PubMed_for_pubs(copy.deepcopy(running_pubs3), config_dict_Hunter_only["Authors"], "asdf", original_queries["PubMed"])
running_pubs5, _ = search_ORCID_for_pubs(copy.deepcopy(running_pubs4), "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
running_pubs6, _ = search_Crossref_for_pubs(copy.deepcopy(running_pubs5), config_dict_Hunter_only["Authors"], "asdf", original_queries["Crossref"])

with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_Google_Scholar", "running_pubs1.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs1, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_Google_Scholar", "running_pubs2.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs2, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_Google_Scholar", "running_pubs3.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs3, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_Google_Scholar", "running_pubs4.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs4, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_Google_Scholar", "running_pubs5.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs5, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_Google_Scholar", "running_pubs6.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs6, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_Google_Scholar", "publication_dict.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs6, indent=2, sort_keys=True))



running_pubs = {}
running_pubs1, _ = search_ORCID_for_pubs(running_pubs, "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
running_pubs2, _ = search_Google_Scholar_for_pubs(copy.deepcopy(running_pubs1), config_dict_Hunter_only["Authors"], "asdf", original_queries["Google Scholar"])
running_pubs3, _ = search_Crossref_for_pubs(copy.deepcopy(running_pubs2), config_dict_Hunter_only["Authors"], "asdf", original_queries["Crossref"])

running_pubs4, _ = search_ORCID_for_pubs(copy.deepcopy(running_pubs3), "asdf", "asdf", config_dict_Hunter_only["Authors"], original_queries["ORCID"])
running_pubs5, _ = search_Google_Scholar_for_pubs(copy.deepcopy(running_pubs4), config_dict_Hunter_only["Authors"], "asdf", original_queries["Google Scholar"])
running_pubs6, _ = search_Crossref_for_pubs(copy.deepcopy(running_pubs5), config_dict_Hunter_only["Authors"], "asdf", original_queries["Crossref"])

with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_PubMed", "running_pubs1.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs1, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_PubMed", "running_pubs2.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs2, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_PubMed", "running_pubs3.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs3, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_PubMed", "running_pubs4.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs4, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_PubMed", "running_pubs5.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs5, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_PubMed", "running_pubs6.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs6, indent=2, sort_keys=True))
with open(os.path.join("tests", "testing_files", "new_intermediate_results", "author_search", "no_PubMed", "publication_dict.json"),'w') as jsonFile:
    jsonFile.write(json.dumps(running_pubs6, indent=2, sort_keys=True))









# -*- coding: utf-8 -*-


import re
import os
import json
import copy

import pymed
import pytest
import xml.etree.ElementTree as ET 

from academic_tracker.fileio import load_json
from academic_tracker import __main__
from academic_tracker.helper_functions import vprint, regex_match_return, regex_group_return, regex_search_return
from academic_tracker.helper_functions import match_pub_authors_to_config_authors, match_pub_authors_to_citation_authors, match_authors_in_prev_pub
from academic_tracker.helper_functions import create_pub_dict_for_saving_PubMed, is_fuzzy_match_to_list, fuzzy_matches_to_list, is_pub_in_publication_dict 
from academic_tracker.helper_functions import create_authors_by_project_dict, adjust_author_attributes, find_duplicate_citations, are_citations_in_pub_dict
from fixtures import publication_dict, pub_with_grants, pub_with_matching_author, passing_config, authors_by_project_dict



def test_vprint(capsys):
    
    vprint("asdf")
    captured = capsys.readouterr()
    
    assert captured.out == "asdf\n"
    

def test_vprint_silent(monkeypatch, capsys):
    monkeypatch.setattr(__main__, "SILENT", True)
    
    vprint("asdf")
    captured = capsys.readouterr()
    
    assert captured.out == ""
    
    
def test_vprint_verbose_on(capsys):
    
    vprint("asdf", verbosity=1)
    captured = capsys.readouterr()
    
    assert captured.out == "asdf\n"
    

def test_vprint_verbose_off(monkeypatch, capsys):
    monkeypatch.setattr(__main__, "VERBOSE", False)
    
    vprint("asdf", verbosity=1)
    captured = capsys.readouterr()
    
    assert captured.out == ""


@pytest.mark.parametrize("regex, string_to_match, return_value", [
        
        (r"(?i).*doi:\s*([^\s]+\w).*", "asdfasdfasdf", ()),
        (r"(?i).*doi:\s*([^\s]+\w).*", "doi: asdf", re.match(r"(?i).*doi:\s*([^\s]+\w).*", "doi: asdf").groups()),
        ]) 
    
def test_regex_match_return(regex, string_to_match, return_value):
    
    assert regex_match_return(regex, string_to_match) == return_value



@pytest.mark.parametrize("regex_group, index, return_value", [
        
        (re.match(r"(?i).*doi:\s*([^\s]+\w).*", "doi: asdf").groups(), 0, "asdf"),
        ((), 0, ""),
        (re.match(r"(a*)(b*)", "aaaaabb").groups(), 1, "bb")
        ]) 
    
def test_regex_group_return(regex_group, index, return_value):
    
    assert regex_group_return(regex_group, index) == return_value



@pytest.mark.parametrize("regex, string_to_match, return_value", [
        
        (r"(?i).*doi:\s*([^\s]+\w).*", "asdfasdfasdf", ()),
        (r"(?i).*doi:\s*([^\s]+\w).*", "doi: asdf", re.search(r"(?i).*doi:\s*([^\s]+\w).*", "doi: asdf").groups()),
        ]) 
    
def test_regex_search_return(regex, string_to_match, return_value):
    
    assert regex_search_return(regex, string_to_match) == return_value



@pytest.fixture
def authors_json_file():
    return {
        "Isabel Escobar": {
              "ORCID": "0000-0001-9269-5927",
              "affiliations": [
                "kentucky"
              ],
              "cutoff_year": 2020,
              "email": "isabel.escobar@uky.edu",
              "first_name": "Isabel",
              "last_name": "Escobar",
              "pubmed_name_search": "Isabel Escobar",
              "scholar_id": "RfB5L8kAAAAJ"
            },
        "Hunter Moseley": {
            "ORCID": "0000-0003-3995-5368",
            "affiliations": [
              "kentucky"
            ],
            "cutoff_year": 2020,
            "email": "hunter.moseley@gmail.com",
            "first_name": "Hunter",
            "last_name": "Moseley",
            "pubmed_name_search": "Hunter Moseley",
            "scholar_id": "ctE_FZMAAAAJ"
          }}
    
    
@pytest.mark.parametrize("PM_author_list, returned_PM_author_list", [
        
        ([{ 'affiliation': 'Department of Neurology, University Medical Center, Johannes Gutenberg-University, Mainz, Germany.',
          'firstname': 'Carine',
          'initials': 'C',
          'lastname': 'Thalman',
          'ORCID':None},
        { 'affiliation': 'Department of Biostats, Kentucky',
          'firstname': 'Hunter',
          'initials': 'HM',
          'lastname': 'Moseley',
          'ORCID':None},],
        [{ 'affiliation': 'Department of Neurology, University Medical Center, Johannes Gutenberg-University, Mainz, Germany.',
          'firstname': 'Carine',
          'initials': 'C',
          'lastname': 'Thalman',
          'ORCID':None},
        { 'affiliation': 'Department of Biostats, Kentucky',
          'firstname': 'Hunter',
          'initials': 'HM',
          'lastname': 'Moseley',
          'author_id': 'Hunter Moseley',
          'ORCID':"0000-0003-3995-5368"},]),
           
        ([{ 'affiliation': 'Department of Neurology, University Medical Center, Johannes Gutenberg-University, Mainz, Germany.',
          'firstname': 'Carine',
          'initials': 'C',
          'lastname': 'Thalman',
          'ORCID':None}], 
            []),
        
        ([{ 'affiliation': 'Department of Biostats, Kentucky',
          'firstname': 'Hunter N. B.',
          'initials': 'HM',
          'lastname': 'Moseley',
          'ORCID':None},],
        [{ 'affiliation': 'Department of Biostats, Kentucky',
          'firstname': 'Hunter N. B.',
          'initials': 'HM',
          'lastname': 'Moseley',
          'author_id': 'Hunter Moseley',
          'ORCID':"0000-0003-3995-5368"},]),
        
        ([{ 'affiliation': 'Department of Biostats, Kentucky',
          'firstname': 'X Hunter',
          'initials': 'HM',
          'lastname': 'Moseley',
          'ORCID':None},],
        [{ 'affiliation': 'Department of Biostats, Kentucky',
          'firstname': 'X Hunter',
          'initials': 'HM',
          'lastname': 'Moseley',
          'author_id': 'Hunter Moseley',
          'ORCID':"0000-0003-3995-5368"},]),
        
        ([{ 'collectivename': 'some name',
          'ORCID':None},],
        [{ 'collectivename': 'some name',
          'author_id': 'Hunter Moseley',
          'ORCID':"0000-0003-3995-5368"},]),
        
        ([{ 'affiliation': 'Department of Biostats, Kentucky',
          'firstname': 'Hunter N. B.',
          'initials': 'HM',
          'lastname': None,
          'ORCID':None},],
        []),
        ])
        
        
        
def test_match_pub_authors_to_config_authors(authors_json_file, PM_author_list, returned_PM_author_list):
    if "collectivename" in PM_author_list[0]:
        authors_json_file["Hunter Moseley"]["collective_name"] = "some name"
    
    assert match_pub_authors_to_config_authors(authors_json_file, PM_author_list) == returned_PM_author_list



def test_match_authors_in_pub_PubMed_collective_names():
    
    citation_authors = [
                      {
                        "initials": "J",
                        "last": "Mitchell"
                      },
                      {"collective_name": "some name"},
                      {
                        "initials": "R",
                        "last": "Flight"
                      },
                      {
                        "initials": "H",
                        "last": "Moseley"
                      }
                    ]
    
    publication_authors = [
        {"collectivename": "some name",
          "ORCID": None}
        ]
    
    assert match_pub_authors_to_citation_authors(citation_authors, publication_authors) == True
    
    
def test_match_authors_in_pub_PubMed_ORCID():
    
    citation_authors = [
                      {
                        "initials": "J",
                        "last": "Mitchell"
                      },
                      {"ORCID": "asdf"},
                      {
                        "initials": "R",
                        "last": "Flight"
                      },
                      {
                        "initials": "H",
                        "last": "Moseley"
                      }
                    ]
    
    publication_authors = [
        {"ORCID": "asdf",
          "lastname": "qwer"}
        ]
    
    assert match_pub_authors_to_citation_authors(citation_authors, publication_authors) == True


@pytest.fixture
def pub_no_PMCID():
    xml_path = os.path.join("tests", "testing_files", "no_author.xml")
    tree = ET.parse(xml_path)
    return pymed.article.PubMedArticle(xml_element=tree.getroot())


def test_modify_pub_dict_for_saving_no_PMCID(pub_no_PMCID):
    modified_pub = load_json(os.path.join("tests", "testing_files", "PubMed_modified_to_save_no_PMCID.json"))
            
    _, pub_to_check = create_pub_dict_for_saving_PubMed(pub_no_PMCID)
    
    # with open(os.path.join("tests", "testing_files", "PubMed_modified_to_save_no_PMCID_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(pub_to_check, indent=2, sort_keys=True))
    
    assert pub_to_check == modified_pub



@pytest.fixture
def pub_with_PMCID():
    xml_path = os.path.join("tests", "testing_files", "pub_with_PMCID.xml")
    tree = ET.parse(xml_path)
    return pymed.article.PubMedArticle(xml_element=tree.getroot())


def test_modify_pub_dict_for_saving_with_PMCID(pub_with_PMCID):
    modified_pub = load_json(os.path.join("tests", "testing_files", "PubMed_modified_to_save_with_PMCID.json"))
            
    _, pub_to_check = create_pub_dict_for_saving_PubMed(pub_with_PMCID, True)
    
    # with open(os.path.join("tests", "testing_files", "PubMed_modified_to_save_with_PMCID_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(pub_to_check, indent=2, sort_keys=True))
    
    assert pub_to_check == modified_pub



@pytest.fixture
def modified_PubMed_XML():
    xml_path = os.path.join("tests", "testing_files", "modified_PubMed_XML.xml")
    tree = ET.parse(xml_path)
    return pymed.article.PubMedArticle(xml_element=tree.getroot())


def test_modify_pub_dict_for_saving_rare_cases(modified_PubMed_XML):
    modified_pub = load_json(os.path.join("tests", "testing_files", "PubMed_rare_cases.json"))
            
    _, pub_to_check = create_pub_dict_for_saving_PubMed(modified_PubMed_XML, True)
    
    # with open(os.path.join("tests", "testing_files", "PubMed_rare_cases_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(pub_to_check, indent=2, sort_keys=True))
    
    assert pub_to_check == modified_pub



@pytest.fixture
def collective_author_XML():
    xml_path = os.path.join("tests", "testing_files", "collective_author_XML.xml")
    tree = ET.parse(xml_path)
    return pymed.article.PubMedArticle(xml_element=tree.getroot())


def test_match_authors_in_prev_pub(collective_author_XML):
    _, pub_to_check = create_pub_dict_for_saving_PubMed(collective_author_XML, True)
    new_author_list = copy.deepcopy(pub_to_check['authors'])
    final_list = copy.deepcopy(pub_to_check['authors'])
    for author in new_author_list[0:-1]:
        author['author_id'] = ''
    
    pub_to_check['authors'].append({'author_id':None, 'firstname':None, 'lastname':None, 'ORCID':'some id'})
    new_author_list.append({'author_id':123, 'firstname':None, 'lastname':None, 'ORCID':'some id'})
    final_list.append({'author_id':123, 'firstname':None, 'lastname':None, 'ORCID':'some id'})
    
    new_author_list.append({'author_id':456, 'firstname':'name', 'lastname':'last', 'ORCID':'some id2'})
    final_list.append({'author_id':456, 'firstname':'name', 'lastname':'last', 'ORCID':'some id2'})
    
    pub_to_check['authors'].append({'author_id':None, 'collectivename':'some name', 'ORCID':None})
    new_author_list.append({'author_id':None, 'collectivename':'some name', 'ORCID':None})
    final_list.append({'author_id':None, 'collectivename':'some name', 'ORCID':None})
        
    combined_author_list = match_authors_in_prev_pub(pub_to_check['authors'], new_author_list)    
        
    assert [item in final_list for item in combined_author_list]



@pytest.mark.parametrize("str_to_match, list_to_matched, is_match", [
        
        ("asdf", ["qwer", "asdf", "zxcv"], True),
        ("asdf", ["qwer", "zxcv"], False)
        ]) 

def test_is_fuzzy_match_to_list(str_to_match, list_to_matched, is_match):
    assert is_fuzzy_match_to_list(str_to_match, list_to_matched) == is_match
    

@pytest.mark.parametrize("str_to_match, list_to_matched, matches", [
        
        ("asdf", ["qwer", "asdf", "zxcv"], [(1,"asdf")]),
        ("asdf", ["qwer", "zxcv"], [])
        ]) 

def test_fuzzy_matches_to_list(str_to_match, list_to_matched, matches):
    assert fuzzy_matches_to_list(str_to_match, list_to_matched) == matches


@pytest.mark.parametrize("pub_id, title, titles, is_in_pub_dict", [
        
        ("https://doi.org/10.1016/j.chroma.2021.462426", "asdf", [], True),
        ("asdf", "Direct injection analysis of per and polyfluoroalkyl substances in surface and drinking water by sample filtration and liquid chromatography-tandem mass spectrometry.", 
          ["Direct injection analysis of per and polyfluoroalkyl substances in surface and drinking water by sample filtration and liquid chromatography-tandem mass spectrometry.",          
          "Untargeted Stable Isotope Probing of the Gut Microbiota Metabolome Using 13C-Labeled Dietary Fibers."], True),
        ("asdf", "qwer", [], False)
        ]) 

def test_is_pub_in_publication_dict(pub_id, title, publication_dict, titles, is_in_pub_dict):
    assert is_pub_in_publication_dict(pub_id, title, publication_dict, titles) == is_in_pub_dict

  

def test_create_authors_by_project_dict(passing_config, authors_json_file, authors_by_project_dict, capsys):
    
    authors_by_project_dict_check = create_authors_by_project_dict(passing_config) 
    captured = capsys.readouterr()
    
    assert authors_by_project_dict_check == authors_by_project_dict and captured.out == "Warning: The author, Isabel Escobar, in the project 1 project of the project tracking configuration file could not be found in the Authors section of the Configuration JSON file.\n"
    
    
    
def test_adjust_author_attributes(authors_by_project_dict, passing_config):
    
    authors_by_project_dict["project 1"]["Hunter Moseley"]["affiliations"] = ["asdf"]
    authors_by_project_dict["project 1"]["Hunter Moseley"]["grants"] = ["asdf", "qwer"]
    authors_by_project_dict["project 1"]["Hunter Moseley"]["cutoff_year"] = 2000
    
#    del authors_by_project_dict["project 2"]["Hunter Moseley"]["affiliations"]
#    del authors_by_project_dict["project 2"]["Hunter Moseley"]["grants"]
#    del authors_by_project_dict["project 2"]["Hunter Moseley"]["cutoff_year"]
    
    modified_authors_json_file = {'Andrew Morris': {'ORCID': '0000-0003-1910-4865',
                                      'affiliations': ['kentucky'],
                                      'collaborator_report': {},
                                      'cutoff_year': 2020,
                                      'email': 'a.j.morris@uky.edu',
                                      'first_name': 'Andrew',
                                      'last_name': 'Morris',
                                      'pubmed_name_search': 'Andrew Morris',
                                      'scholar_id': '-j7fxnEAAAAJ',
                                      'grants': ['P42 ES007380', 'P42ES007380']},
                                  'Hunter Moseley': {'ORCID': '0000-0003-3995-5368',
                                      'affiliations': ['asdf', 'kentucky'],
                                      'collaborator_report': {},
                                      'cutoff_year': 2000,
                                      'email': 'hunter.moseley@gmail.com',
                                      'first_name': 'Hunter',
                                      'last_name': 'Moseley',
                                      'pubmed_name_search': 'Hunter Moseley',
                                      'scholar_id': 'ctE_FZMAAAAJ',
                                      'grants': ['P42 ES007380', 'P42ES007380', 'asdf', 'qwer']}}
        
    adjust_author_attributes(authors_by_project_dict, passing_config)
    
    assert passing_config["Authors"] == modified_authors_json_file
    
    


@pytest.fixture
def tokenized_citations():
    return [{"PMID":"1234", "DOI":"ASDF", "title":"made up title"},
            {"PMID":"", "DOI":"ASDF", "title":""},
            {"PMID":"1234", "DOI":"ASDF", "title":""},
            
            {"PMID":"", "DOI":"QWER", "title":""},
            {"PMID":"", "DOI":"qwer", "title":""},
            
            {"PMID":"4567", "DOI":"", "title":""},
            {"PMID":"4567", "DOI":"", "title":""},
            
            {"PMID":"", "DOI":"", "title":"new title"},
            {"PMID":"", "DOI":"", "title":"new titles"},]
    
def test_find_duplicate_citations(tokenized_citations):
    
    duplicate_citations_check = set([(0,1,2), (3,4), (5,6), (7,8)])
    
    duplicate_citations = find_duplicate_citations(tokenized_citations)
    
    duplicate_citations = {tuple(duplicates) for duplicates in duplicate_citations}
    
    assert duplicate_citations == duplicate_citations_check
    
    
    
@pytest.fixture
def publication_json():
    return load_json(os.path.join("tests", "testing_files", "publication_dict.json"))


def test_are_citations_in_pub_dict(publication_json):
    
    tokenized_citations = [{"PMID":"35313030", "DOI":"", "title":""},
                            {"PMID":"", "DOI":"10.1038/s41467-023-35784-x", "title":""},
                            {"PMID":"", "DOI":"", "title":"kegg_pull: a software package for the RESTful access and pulling from the Kyoto Encyclopedia of Gene and Genomes."},
                            {"PMID":"1234", "DOI":"", "title":""}]
    
    is_citation_in_pubs_check = [True, True, True, False]
    
    is_citation_in_pubs = are_citations_in_pub_dict(tokenized_citations, publication_json)
    
    assert is_citation_in_pubs == is_citation_in_pubs_check
    
    
    
